from threading import RLock
from typing import Any, Callable, Dict, Generic, Iterable, Iterator, List, Optional, Set, Tuple, Type, TypeVar, Union


Scheme = type(List)

T = TypeVar("T")
K = TypeVar("K")


def _check_structure(s : Scheme, t : Any) -> bool:
    if s == t:
        return True
    if not hasattr(s, "__origin__") or s.__origin__ not in (list, tuple, set, Union):
        return False
    return all(_check_structure(si, t) for si in s.__args__)



class View(Generic[K, T]):


    def __init__(self, database : "Database", map : Optional[Callable[[T], K]]) -> None:
        if not isinstance(database, Database):
            raise TypeError("Expected Database type for database, got " + repr(database.__class__.__name__))
        if map != None and not callable(map):
            raise TypeError(repr(map.__class__.__name__) + " object is not callable.")
        self.entries : Dict[K, Tuple[List[T], Set[T]]] = {}
        self.database : Database = database
        self.map = map
        self.refresh()
    

    def close(self):
        self.database = None
    

    def __str__(self) -> str:
        if self.entries:
            for k in self.entries:
                break
            return "View[" + type(k).__name__ + "]"
        return "View[]"
    

    def __setattr__(self, name: str, value: Any) -> None:
        if hasattr(self, "database") and self.database == None:
            raise RuntimeError("View has been closed.")
        return super().__setattr__(name, value)
    

    def __delattr__(self, name: str) -> None:
        if name == "database":
            raise AttributeError("Cannot delete reference database.")
        return super().__delattr__(name)
        

    def refresh(self):
        if self.database == None:
            raise RuntimeError("View has been closed.")
        with self.database.lock:
            self.entries = {}
            if self.map != None:
                for t in self.database:
                    try:
                        k = self.map(t)
                    except:
                        from traceback import format_exc
                        raise ValueError("Given mapping function raised an exception for entry {} :\n{}".format(t, format_exc()))
                    if k not in self.entries:
                        try:
                            self.entries[k] = ([], set())
                        except:
                            from traceback import format_exc
                            raise ValueError("Could not use key {} as an entry. It must be hashable :\n{}".format(k, format_exc()))
                    try:
                        self.entries[k][1].add(t)
                    except:
                        self.entries[k][0].append(t)


    def __getitem__(self, i : T) -> K:
        for k, (l, s) in self.entries.items():
            if i in s or i in l:
                return k
        return None
    

    def __setitem__(self, i : T, v : K):
        for k, (l, s) in self.entries.items():
            if i in s:
                s.remove(i)
            elif i in l:
                l.remove(i)
        if v not in self.entries:
            try:
                self.entries[v] = ([], set())
            except:
                raise ValueError("Unhashable type : " + repr(v.__class__.__name__))
        try:
            self.entries[v][1].add(i)
        except:
            self.entries[v][0].append(i)
    

    def values(self):
        return self.entries.keys()
    

    def by_value(self, v : K) -> Iterator[T]:
        if v not in self.entries:
            return iter(())
        from itertools import chain
        return chain(self.entries[v][0], self.entries[v][1])
    

    def by_values(self) -> Iterator[Tuple[K, T]]:
        for k in self.entries:
            for t in self.by_value(k):
                yield k, t





class _Toolbox:
        
        """
        Just a generic function container.
        """



    

class Database(Generic[T]):


    def __init__(self, cls : Type[T], storage_structure : Scheme) -> None:
        if not _check_structure(storage_structure, cls):
            raise TypeError("Structure can only be made of List, Tuple, Set, Union and Target type")
        self.scheme = storage_structure
        self.target = cls
        self.database : Scheme = storage_structure.__origin__()
        self.lock = RLock()
        self.views : Dict[str, View[Any, cls]] = {}
        self.__tools : _Toolbox = _Toolbox()
    

    def clear(self):
        """
        Clears the database.
        """
        with self.lock:
            self.database = self.scheme.__origin__()
            self.views = {}

        
    @property
    def toolbox(self) -> _Toolbox:
        """
        A toolbox full of functions designed to handle this particular database.
        """
        return self.__tools
    

    def __str__(self) -> str:
        return "Database[" + str(self.target) + "]"


    def __iter__(self) -> Iterator[T]:
        def deep_iter(s : Iterable):
            with self.lock:
                for si in s:
                    if isinstance(si, self.target):
                        yield si
                    elif isinstance(si, Iterable):
                        for sii in deep_iter(si):
                            yield sii
        
        return deep_iter(self.database)
    

    def __len__(self) -> int:

        def deep_dumb_iter(s : Iterable):
            if isinstance(s, self.target):
                return 1
            else:
                return sum(deep_dumb_iter(si) for si in s)

        def deep_len(s : Iterable, structure : Scheme):
            with self.lock:
                if structure == self.target:
                    return 1
                if structure.__origin__ in (list, set):
                    if structure.__args__[0] == self.target:
                        return len(s)
                    else:
                        return sum(deep_len(li, structure.__args__[0]) for li in s)
                elif structure.__origin__ == tuple:
                    return sum(deep_len(si, ai) for si, ai in zip(s, structure.__args__))
                else:
                    return deep_dumb_iter(s)
        
        return deep_len(self.database, self.scheme)


    def __dir__(self) -> Iterable[str]:
        return super().__dir__() + list(self.views)
    

    def refresh(self):
        with self.lock:
            for v in self.views.values():
                v.refresh()

    
    def create_view(self, name : str, map : Optional[Callable[[T], K]] = None) -> View[K, T]:
        if not isinstance(name, str):
            raise TypeError("Expected str for view's name, got " + repr(name.__class__.__name__))
        try:
            with self.lock:
                v = View(self, map)
        except:
            raise
        self.views[name] = v
        return v
    

    def remove_view(self, name : str) -> View:
        if name not in self.views:
            raise KeyError("No such view : " + repr(name))
        v = self.views.pop(name)
        v.close()
        return v
    

    def __getattribute__(self, name: str) -> Any:
        try:
            return super().__getattribute__(name)
        except AttributeError:
            if name in self.views:
                return self.views[name]
            raise
    

    def __reduce__(self) -> str | tuple[Any, ...]:
        return Database, (self.target, self.scheme), (self.database, self.views, self.__tools)
    
    
    def __setstate__(self, state : Tuple[Scheme, Dict[str, View], toolbox]):
        self.database, self.views, self.__tools = state



"""
This is a *magic* script to bake baguettes. Use -h/--help for usage.
"""





def main():
    """
    Command line function to bake Cuckoo report(s). Use -h/--help for more info.
    """

    if __name__ in ("__main__", "bake"):        # One for when started with Python, the other for when started with package command...

        import logging
        from baguette.bakery.logger import logger, set_level
        set_level(logging.ERROR)

        import argparse
        from os import environ
        import pathlib
        from typing import Literal, Iterator
        from baguette.bakery.compiler import compile
        from baguette.bakery.source import filters
        from baguette.rack import BaguetteRack, TimeoutExit

        parser = argparse.ArgumentParser(
            description = 'Bakes Cuckoo reports into baguettes (gexf and pyt graphs).',
            add_help = True,
            conflict_handler = 'resolve',
            epilog = """
            Note that the format of input and outputs is quite flexible. Indeed you can :
            give output folders for each output file, which will make all input have their output in these folders
            - give a set of outputs for each input file, as long as they are given in the same place in the parameter sequences
            - do the same with input folders and output folders, as long as they also come in the same order
            - mix these options, again, as long as it has a meaning in the order they come.
            - Note that if a given path does not exists, it is interpreted as a folder.
            - Also, if an input path has the appropriate extension, it will be considered as a single input.
            """
            )

        class PathSorter:

            def __init__(self, pool : list[pathlib.Path | Literal["-"]], name : str) -> None:
                self.pool = pool
                self.name = name
            
            def __call__(self, arg : str):
                if arg == "-":
                    self.pool.append(arg)
                    return
                from glob import iglob
                import re
                magic_check = re.compile('([*?[])')
                try:
                    if magic_check.search(arg) is not None:
                        self.pool.extend(pathlib.Path(path) for path in iglob(arg))
                    else:
                        self.pool.append(pathlib.Path(arg))
                except:
                    parser.error("invalid {} path : '{}'".format(self.name, arg))

        reports : list[pathlib.Path | Literal["-"]] = []
        baguettes : list[pathlib.Path | Literal["-"]] = []
        visuals : list[pathlib.Path | Literal["-"]] = []
        outputs : list[pathlib.Path | Literal["-"]] = []

        def pool_size(arg : str) -> int:
            """
            Transforms a numeric argument in a number of process to use as a process pool.
            It can be absolute, negative (relative to the number of CPUs) or 
            """
            from os import cpu_count
            N = cpu_count()
            if not N:
                N = 1
            try:
                v = int(arg)
                if v < 0:
                    v = N - v
                if v <= 0:
                    parser.error("got a (too) negative value for process pool size : '{}'".format(arg))
            except:
                try:
                    v = float(arg)
                    if v <= 0:
                        parser.error("got a negative relative process pool size : '{}'".format(arg))
                    v = round(v * N)
                except:
                    parser.error("not a process pool size : '{}'".format(arg))
            return v

        def time(arg : str) -> float:
            from math import isnan
            try:
                v = float(arg)
                if v <= 0 or isnan(v):
                    parser.error("got a negative, null or nan maxtime")
                return v
            except:
                parser.error("expected positive float for maxtime, got : '{}'".format(arg))

        parser.add_argument("reports", type=PathSorter(reports, "report"), default=None, nargs="*" if "BAGUETTE_REPORTS" in environ else "+", help="Cuckoo report files (.json) to bake baguettes from. Can also be folders containing reports. Defaults to environment variable 'BAGUETTE_REPORTS' if set.")
        parser.add_argument("--baguettes", type=PathSorter(baguettes, "baguette"), default=None, action="extend", nargs="*", help="the path(s) to the output baguette files (.pyt). Use '-' to leave it to the automatic destination. Defaults to environment variable 'BAGUETTE_BAGUETTES' or '-' if not set.")
        parser.add_argument("--visuals", type=PathSorter(visuals, "visual"), default=None, action="extend", nargs="*", help="the path(s) to the output visual (Gephi) files (.gexf). Use '-' to leave it to the automatic destination. Defaults to environment variable 'BAGUETTE_VISUALS' or '-' if not set.")
        parser.add_argument("-o", "--outputs", type=PathSorter(outputs, "output"), default=None, action="extend", nargs="*", help="the path to the result index folders (which end in .bag). They contain the index file (.pyt) which stores all the information about a given baguette. Use '-' to leave it to the automatic destination. Defaults to environment variable 'BAGUETTE_OUTPUTS' or '.' if not set.")
        parser.add_argument("--pool", type=pool_size, default=pool_size("0.5"), help="the size of the process pool to use to bake in parallel.")
        parser.add_argument("--maxtime", type=time, default=time("inf"), help="the maximum amount of time spent baking a single baguette. No maxtime by default.")
        parser.add_argument("-f", "--filters", type=str, default=[], choices=[name for name in dir(filters) if isinstance(getattr(filters, name), filters.Filter)], nargs="*", help="a list of filters that can be used when exporting the baguette to the visual file (.gexf).")
        parser.add_argument("--perf", action="store_true", default=False, help="if this is enabled, a performance report will be printed at the end of the baking process.")
        parser.add_argument("-v", "--verbosity", action="count", default=0, help="increases the verbosity of the output.")
        parser.add_argument("--skip_data_comparison", action="store_true", default=False, help="if enabled, the computation of the Levenshtein similarity between all Data nodes will be skipped.")
        parser.add_argument("--skip_diff_comparison", action="store_true", default=False, help="same as skip_data_comparison but for Diff nodes.")

        args = parser.parse_args()

        # Setting logging level

        levels = {
            0 : logging.ERROR,
            1 : logging.WARNING,
            2 : logging.INFO,
            3 : logging.DEBUG
        }
        verbosity : Literal[0, 1, 2, 3] = min(3, args.verbosity)
        set_level(levels[verbosity])

        logger.info("Arguments parsed. Discovering jobs.")

        # Parsing jobs

        if not reports:
            try:
                reports = [pathlib.Path(environ["BAGUETTE_REPORTS"])]
            except KeyError:
                raise RuntimeError("Environment variable 'BAGUETTE_REPORTS' not found and there is no input...")
            except:
                parser.error("invalid report path in environment variable : '{}'".format(environ["BAGUETTE_REPORTS"]))
        if not baguettes:
            if "BAGUETTE_BAGUETTES" in environ:
                try:
                    baguettes = [pathlib.Path(environ["BAGUETTE_BAGUETTES"])]
                except:
                    parser.error("invalid baguettes path in environment variable : '{}'".format(environ["BAGUETTE_BAGUETTES"]))
            else:
                baguettes = ["-"]
        if not visuals:
            if "BAGUETTE_VISUALS" in environ:
                try:
                    visuals = [pathlib.Path(environ["BAGUETTE_VISUALS"])]
                except:
                    parser.error("invalid visuals path in environment variable : '{}'".format(environ["BAGUETTE_VISUALS"]))
            else:
                visuals = ["-"]
        if not outputs:
            if "BAGUETTE_OUTPUTS" in environ:
                try:
                    outputs = [pathlib.Path(environ["BAGUETTE_OUTPUTS"])]
                except:
                    parser.error("invalid outputs path in environment variable : '{}'".format(environ["BAGUETTE_OUTPUTS"]))
            else:
                outputs = [pathlib.Path(".")]

        class JobQueue:

            """
            These objects hold a series of grouped jobs.
            """

            def __init__(self, groups : list[pathlib.Path | Literal["-"] | list[pathlib.Path]] = []) -> None:
                from copy import deepcopy
                self.__groups : list[pathlib.Path | Literal["-"] | list[pathlib.Path]] = deepcopy(groups)
                self.__last_group : list[pathlib.Path] = []
                self.__active : bool = False
            
            def insert_group(self, group : list[pathlib.Path] | pathlib.Path | Literal["-"]):
                """
                Creates an independant work group with given content, also creating a new active group for later use.
                """
                if self.__active:
                    self.__groups.append(self.__last_group)
                self.__groups.append(group)
                self.__last_group = []
                self.__active = False
            
            def append_to_active_group(self, path : pathlib.Path):
                """
                Appends a file to the current group.
                """
                self.__last_group.append(path)
                self.__active = True
            
            def new_group(self):
                """
                Adds the active group to the queue and creates a new empty active group.
                """
                if self.__active:
                    self.__groups.append(self.__last_group)
                self.__last_group = []
                self.__active = True

            def finalize(self):
                """
                Finalizes the queue, adding the currently active group to the queue if necessary.
                """
                if self.__active and self.__last_group:
                    self.__groups.append(self.__last_group)
            
            def __iter__(self) -> Iterator[pathlib.Path | Literal["-"] | list[pathlib.Path]]:
                """
                Yields all the jobs in the queue.
                """
                yield from self.__groups
            
            def __len__(self) -> int:
                """
                Implements len(self).
                """
                return len(self.__groups)

            def __getitem__(self, index : int) -> pathlib.Path | Literal["-"] | list[pathlib.Path]:
                """
                Implements self[index].
                """
                return self.__groups[index]


        report_groups : JobQueue = JobQueue()
        baguette_groups : JobQueue = JobQueue()
        visual_groups : JobQueue = JobQueue()
        output_groups : JobQueue = JobQueue()

        def determine_type(path : pathlib.Path | Literal["-"], ext : str) -> Literal["folder", "file"]:
            """
            Determines if the given path should be interpreted as a folder or file path.
            If it ends with the given extension and does not exist, it well be considered a file.
            """
            if path == "-":
                return "folder"
            elif ext and path.suffix == ext:
                return "file"
            if path.is_file():
                return "file"
            elif path.is_dir():
                return "folder"
            elif path.exists():
                parser.error("given path exists and is neither a file or folder : '{}'".format(path))
            else:
                return "folder"
        
        def create_name(folder : pathlib.Path) -> pathlib.Path:
            """
            Given a path, this will return a (possibly) modified path that does not exist in the same folder.
            """
            if not folder.exists():
                return folder
            n = 0
            ext = folder.suffix
            if ext:
                sfolder = str(folder)[:-len(ext)]
            else:
                sfolder = str(folder)
            new_folder = pathlib.Path(sfolder + "({})".format(n) + ext)
            while new_folder.exists():
                n += 1
                new_folder = pathlib.Path(sfolder + "({})".format(n) + ext)
            return new_folder
            

        for r in reports:
            if r == "-":
                parser.error("cannot use automatic destination operator for input files.")
            if determine_type(r, ".json") == "folder":
                if r.exists():
                    report_groups.insert_group([p for p in r.iterdir()])
                else:
                    parser.error("input folder/file does not exist : '{}'".format(r))
            else:
                if r.exists():
                    report_groups.append_to_active_group(r)
                else:
                    parser.error("input file does not exist : '{}'".format(r))
        report_groups.finalize()
        
        for b in baguettes:
            if determine_type(b, ".pyt") == "folder":
                baguette_groups.insert_group(b)
            else:
                if b == "-":
                    raise RuntimeError("How?")
                baguette_groups.append_to_active_group(b)
        baguette_groups.finalize()
        if len(baguette_groups) == 1 and len(report_groups) > 1 and (isinstance(baguette_groups[0], pathlib.Path) or baguette_groups[0] == "-"):
            baguette_groups = JobQueue([baguette_groups[0]] * len(report_groups))

        for v in visuals:
            if determine_type(v, ".gexf") == "folder":
                visual_groups.insert_group(v)
            else:
                if v == "-":
                    raise RuntimeError("How?")
                visual_groups.append_to_active_group(v)
        visual_groups.finalize()
        if len(visual_groups) == 1 and len(report_groups) > 1 and (isinstance(visual_groups[0], pathlib.Path) or visual_groups[0] == "-"):
            visual_groups = JobQueue([visual_groups[0]] * len(report_groups))
        
        for o in outputs:
            if determine_type(o, ".bag") == "folder":
                output_groups.insert_group(o)
            else:
                if o == "-":
                    raise RuntimeError("How?")
                output_groups.append_to_active_group(o)
        output_groups.finalize()
        if len(output_groups) == 1 and len(report_groups) > 1 and (isinstance(output_groups[0], pathlib.Path) or output_groups[0] == "-"):
            output_groups = JobQueue([output_groups[0]] * len(report_groups))

        # def printable_format(l) -> str:
        #     return str(l) if not isinstance(l, list) else (str(l) if len(l) < 2 else "[...]")

        # print("Sizes: {}, {}, {}, {}".format(len(report_groups), len(baguette_groups), len(visual_groups), len(output_groups)))
        # print(list(printable_format(r) for r in report_groups))
        # print(list(printable_format(b) for b in baguette_groups))
        # print(list(printable_format(v) for v in visual_groups))
        # print(list(printable_format(o) for o in output_groups))
        # l1 = list(report_groups)
        # l2 = list(baguette_groups) + [None] * (len(report_groups) - len(baguette_groups))
        # l3 = list(visual_groups) + [None] * (len(report_groups) - len(visual_groups))
        # l4 = list(output_groups) + [None] * (len(report_groups) - len(output_groups))
        # for r, b, v, o in zip(l1, l2, l3, l4):
        #     print("Work group:")
        #     print("reports :", printable_format(r))
        #     print("baguettes :", printable_format(b))
        #     print("visuals :", printable_format(v))
        #     print("outputs :", printable_format(o))
        #     print("\n")

        if len(report_groups) != len(baguette_groups) or len(report_groups) != len(visual_groups) or len(report_groups) != len(output_groups):
            parser.error("different number of work groups in inputs/outputs.")
        
        for r, b, v, o in zip(report_groups, baguette_groups, visual_groups, output_groups):
            if not isinstance(r, list):
                raise RuntimeError("How did we get here?")
            n = len(r)
            if isinstance(b, list) and len(b) != n:
                parser.error("got a work group with different numbers of inputs and outputs.")
            if isinstance(v, list) and len(v) != n:
                parser.error("got a work group with different numbers of inputs and outputs.")
            if isinstance(o, list) and len(o) != n:
                parser.error("got a work group with different numbers of inputs and outputs.")

        work : list[BaguetteRack] = []

        def stripname(p : pathlib.Path) -> str:
            """
            Returns the name of the path without the exetension if it has one.
            """
            ext = p.suffix
            if ext:
                return p.name[:-len(ext)]
            return p.name

        for rgroup, bgroup, vgroup, ogroup in zip(report_groups, baguette_groups, visual_groups, output_groups):

            if rgroup == "-":
                parser.error("cannot use automatic destination operator for input files.")
            elif isinstance(rgroup, pathlib.Path):      # Folder input
                rgroup = list(rgroup.iterdir())         # Transform in file inputs

            if ogroup == "-":
                parser.error("cannot use automatic destination operator for output folders.")
            elif isinstance(ogroup, pathlib.Path):      # Folder output
                ogroup = [create_name(pathlib.Path(ogroup, stripname(r) + ".bag")) for r in rgroup]
            
            if bgroup == "-":                           # Magic destination:
                bgroup = [None for _ in rgroup]         # To be determined later
            elif isinstance(bgroup, pathlib.Path):      # Folder baguette destination
                bgroup = [create_name(pathlib.Path(bgroup, stripname(r) + ".pyt")) for r in rgroup]       # Pre-compute the baguette paths with report names

            if vgroup == "-":                           # Magic destination:
                vgroup = [None for _ in rgroup]         # To be determined later
            elif isinstance(vgroup, pathlib.Path):      # Folder visual destination
                vgroup = [create_name(pathlib.Path(vgroup, stripname(r) + ".gexf")) for r in rgroup]       # Pre-compute the visual paths with report names
            
            for r, b, v, o in zip(rgroup, bgroup, vgroup, ogroup):
                bg = BaguetteRack(o)
                bg.report = r
                bg.baguette = b if b is not None else (bg.working_directory / "baguette.pyt")
                bg.visual = v if v is not None else (bg.working_directory / "visual.gexf")
                bg.verbosity = verbosity
                bg.skip_data_comparison = args.skip_data_comparison
                bg.skip_diff_comparison = args.skip_diff_comparison
                bg.filter_names = args.filters
                bg.maxtime = args.maxtime
                bg.perf = args.perf
                work.append(bg)
        
        
        # Compile now...

        from multiprocessing.pool import Pool
        from threading import Lock, Thread

        # All of this is because multiprocessing was coded with feet... Pool's async methods may freeze (deadlock maybe) on some platforms.

        lock = Lock()
        failed, timed_out, total = 0, 0, len(work)
        def execute_single_job(P : Pool) -> bool:
            nonlocal failed, timed_out
            with lock:
                if not work: 
                    return False
                br = work.pop()
            try:
                br = P.apply(compile, (br, ))
            except KeyboardInterrupt as e:
                from traceback import TracebackException
                br.exception = TracebackException.from_exception(e)
            br.export()
            if br.exception is not None and issubclass(br.exception.exc_type, KeyboardInterrupt):
                return False
            elif br.exception is not None and not issubclass(br.exception.exc_type, TimeoutExit):
                with lock:
                    failed += 1
                logger.error("Got a '{}' error during the baking of '{}'.".format(br.exception.exc_type.__name__, br.report.name))
            elif br.exception is not None and issubclass(br.exception.exc_type, TimeoutExit):
                with lock:
                    timed_out += 1
            return True

        def executor(P : Pool):
            while execute_single_job(P):
                pass
        
        threads : list[Thread] = []
        try:
            with Pool(args.pool, maxtasksperchild = 1) as P:
                for _ in range(args.pool):
                    t = Thread(target = executor, args = (P, ), daemon = True)
                    t.start()
                    threads.append(t)

                for t in threads:
                    t.join()
            
            success = total - failed - timed_out
            if failed and success and timed_out:
                print("{} failed baguettes, {} took too long and {} well-baked.".format(failed, timed_out, success))
            elif failed and success:
                print("{} failed baguettes, {} baked correctly.".format(failed, success))
            elif timed_out and success:
                print("{} baguettes took too long to bake, {} baked correctly.".format(timed_out, success))
            elif failed and timed_out:
                print("{} baguettes are failed and the {} others took too long to bake...".format(failed, timed_out))
            elif failed:
                print("All {} baguettes did not bake correctly...".format(failed))
            elif timed_out:
                print("All {} baguettes took too long to bake...".format(timed_out))
            elif success:
                print("All {} are well-baked!".format(success))
        except KeyboardInterrupt:
            print("Exiting.")
            




if __name__ == "__main__":
    main()


"""
This module defines the Vertices and Edges that are related to importations.
"""

from ...logger import logger
from typing import Optional
from ..colors import Color
from ..graph import Edge, UniqueVertex, Vertex
from .filesystem import NewFile, File
from .execution import Process
from ..utils import chrono

__all__ = ["Import", "HasImport", "IsFile"]





logger.debug("Loading imports library.")

class Import(UniqueVertex):
    
    """
    An import vertex. Shows the use of a DLL or equivalent.
    """

    __slots__ = {
        "__name" : "The name of the import file.",
        "path" : "The path to the file.",
        "length" : "The size in bytes of the file."
    }

    __pickle_slots__ = {
        "name",
        "path",
        "length"
    }

    __suspicious_keyword_names = {
        "crypt",
        "advapi",
        "kernel",
        "sock"
    }

    def __init__(self, *, c: Color | None = None, parent: Optional[Vertex] = None) -> None:
        super().__init__(c=c, parent=parent)
        self.__name : str = ""
        self.path : str = ""
        self.length : int = 0

    @property
    def name(self) -> str:
        """
        The name of the imported library.
        """
        return self.__name

    @name.setter
    def name(self, n : str):
        from ..config import ColorMap
        if not isinstance(n, str):
            raise TypeError("Expected str, got " + repr(type(n).__name__))
        self.__name = n
        self.color = ColorMap.get_config().imports.Import
        for kw in self.__suspicious_keyword_names:
            if kw in n.lower():
                self.color = ColorMap.get_config().imports.SuspiciousImport
                break
    
    @property
    def label(self) -> str:
        """
        Returns a label for this Import node.
        """
        return "Import {}".format(self.__name.lower())
    




class HasImport(Edge):

    """
    This kind of edge indicates that a process imported something.
    """

    label : str = ""

    source : Process
    destination : Import





class IsFile(Edge):

    """
    This kind of edge indicates that an import corresponds to an existing file object.
    """

    label : str = ""

    source : Import
    destination : File





@chrono
def link(e : NewFile):
    from ..graph import find_or_create
    if e.file.path.suffix.lower().endswith(".dll"):
        I, ok = find_or_create(Import, path = e.file.name)
        if not ok:
            import re
            expr1 = re.compile(r"([a-zA-Z0-9_.]+)(\..*?)$")
            match = expr1.search(e.file.name)
            if match:
                I.name = match.group(1).lower()
            else:
                I.name = e.file.name
        IsFile(I, e.file)

NewFile.add_callback(link)


del logger, Optional, Color, UniqueVertex, Vertex, NewFile, chrono, link


"""
This module defines the Vertices and Edges that are related to network activity.
"""

from collections import defaultdict
from ...logger import logger
from typing import Any, Optional
from ..colors import Color
from ..graph import Arrow, Edge, UniqueVertex, Vertex
from .execution import Call, CallHandler, Process
from ..utils import chrono
from Viper.meta.decorators import staticproperty

__all__ = ["Socket", "Connection", "Host", "SpawnedProcess", "HasConnection", "HasSocket", "Communicates", "CreatesSocket", "Binds", "Connects", "Sends", "Receives", "Conveys", "Closes", "CloseSocket", "ListensOn", "Shutdown", "Accepts", "Duplicates"]





logger.debug("Loading network library.")

class Connection(UniqueVertex):

    """
    A connection vertex. This is a link vertex between two socket. It might only exist for a part of the lifetime of both its connected sockets.
    """

    __slots__ = {
        "duration" : "The duration that the connection was maintained for",
        "volume" : "The amount of data transfered through the connection",
        "src" : "The source address",
        "dest" : "The destination address"
    }

    __pickle_slots__ = {
        "duration",
        "volume",
        "src",
        "dest"
    }

    def __init__(self, *, c: Color | None = None, parent: Optional[Vertex] = None) -> None:
        super().__init__(c=c, parent=parent)
        self.duration : float = 0.
        self.volume : int = 0
        self.src : Any = None
        self.dest : Any = None

    @property
    def socket(self) -> "Socket":
        """
        The local Socket that this connection was made through.
        """
        for e in self.edges:
            if isinstance(e, HasConnection):
                return e.source
        raise RuntimeError("Got a Connection bound to no Socket.")





class Socket(UniqueVertex):

    """
    A socket vertex. Represents a handle given by the system to allow some kind of communication.
    """

    __slots__ = {
        "family" : "The socket address family. For example, 'InterNetwork' is for IP V4 addresses. Refer to Socket.families for documentation",
        "protocol" : "The transport protocol used by the socket. For example, Tcp. Refer to Socket.protocols for documentation",
        "type" : "The type of socket. For exemple, Dgram is a socket that supports datagrams. Refer to Socket.types for documentation"
    }

    __pickle_slots__ = {
        "family",
        "protocol",
        "type"
    }

    families = defaultdict(lambda : ("Uncharted", "The value does not correspond to any known address family"), {
        16 : ('AppleTalk', 'AppleTalk address.'),
        22 : ('Atm', 'Native ATM services address.'),
        21 : ('Banyan', 'Banyan address.'),
        10 : ('Ccitt', 'Addresses for CCITT protocols, such as X.25.'),
        5 : ('Chaos', 'Address for MIT CHAOS protocols.'),
        24 : ('Cluster', 'Address for Microsoft cluster products.'),
        65537 : ('ControllerAreaNetwork', 'Controller Area Network address.'),
        9 : ('DataKit', 'Address for Datakit protocols.'),
        13 : ('DataLink', 'Direct data-link interface address.'),
        12 : ('DecNet', 'DECnet address.'),
        8 : ('Ecma', 'European Computer Manufacturers Association (ECMA) address.'),
        19 : ('FireFox', 'FireFox address.'),
        15 : ('HyperChannel', 'NSC Hyperchannel address.'),
        1284425 : ('Ieee', 'IEEE 1284.4 workgroup address.'),
        3 : ('ImpLink', 'ARPANET IMP address.'),
        2 : ('InterNetwork', 'Address for IP version 4.'),
        623 : ('InterNetworkV', 'Address for IP version 6.'),
        6 : ('Ipx', 'IPX or SPX address.'),
        26 : ('Irda', 'IrDA address.'),
        7 : ('Iso', 'Address for ISO protocols.'),
        14 : ('Lat', 'LAT address.'),
        29 : ('Max', 'MAX address.'),
        17 : ('NetBios', 'NetBios address.'),
        28 : ('NetworkDesigners', 'Address for Network Designers OSI gateway-enabled protocols.'),
        6 : ('NS', 'Address for Xerox NS protocols.'),
        7 : ('Osi', 'Address for OSI protocols.'),
        65536 : ('Packet', 'Low-level Packet address.'),
        4 : ('Pup', 'Address for PUP protocols.'),
        11 : ('Sna', 'IBM SNA address.'),
        1 : ('Unix', 'Unix local to host address.'),
        -1 : ('Unknown', 'Unknown address family.'),
        0 : ('Unspecified', 'Unspecified address family.'),
        18 : ('VoiceView', 'VoiceView address.'),
    })

    protocols = defaultdict(lambda : ("Uncharted", "The value does not correspond to any known protocol"), {
        3 : ('Ggp', 'Gateway To Gateway Protocol.'),
        1 : ('Icmp', 'Internet Control Message Protocol.'),
        58 : ('IcmpV6', 'Internet Control Message Protocol for IPv6.'),
        22 : ('Idp', 'Internet Datagram Protocol.'),
        2 : ('Igmp', 'Internet Group Management Protocol.'),
        0 : ('IP', 'Internet Protocol.'),
        51 : ('IPSecAuthenticationHeader', 'IPv6 Authentication header. For details'),
        50 : ('IPSecEncapsulatingSecurityPayload', 'IPv6 Encapsulating Security Payload header.'),
        4 : ('IPv4', 'Internet Protocol version 4.'),
        41 : ('IPv6', 'Internet Protocol version 6 (IPv6).'),
        60 : ('IPv6DestinationOptions', 'IPv6 Destination Options header.'),
        44 : ('IPv6FragmentHeader', 'IPv6 Fragment header.'),
        0 : ('IPv6HopByHopOptions', 'IPv6 Hop by Hop Options header.'),
        59 : ('IPv6NoNextHeader', 'IPv6 No next header.'),
        43 : ('IPv6RoutingHeader', 'IPv6 Routing header.'),
        1000 : ('Ipx', 'Internet Packet Exchange Protocol.'),
        77 : ('ND', 'Net Disk Protocol (unofficial).'),
        12 : ('Pup', 'PARC Universal Packet Protocol.'),
        255 : ('Raw', 'Raw IP packet protocol.'),
        1256 : ('Spx', 'Sequenced Packet Exchange protocol.'),
        1257 : ('SpxII', 'Sequenced Packet Exchange version 2 protocol.'),
        6 : ('Tcp', 'Transmission Control Protocol.'),
        17 : ('Udp', 'User Datagram Protocol.'),
        -1 : ('Unknown', 'Unknown protocol.'),
        0 : ('Unspecified', 'Unspecified protocol.'),
    })

    types = defaultdict(lambda : ("Uncharted", "The value does not correspond to any known socket type"), {
        2 : ('Dgram', 'Supports datagrams, which are connectionless, unreliable messages of a fixed (typically small) maximum length. Messages might be lost or duplicated and might arrive out of order. A Socket of type Dgram requires no connection prior to sending and receiving data, and can communicate with multiple peers. Dgram uses the Datagram Protocol (ProtocolType.Udp) and the AddressFamily. InterNetwork address family.'),
        3 : ('Raw', 'Supports access to the underlying transport protocol. Using Raw, you can communicate using protocols like Internet Control Message Protocol (ProtocolType.Icmp) and Internet Group Management Protocol (ProtocolType.Igmp). Your application must provide a complete IP header when sending. Received datagrams return with the IP header and options intact.'),
        4 : ('Rdm', 'Supports connectionless, message-oriented, reliably delivered messages, and preserves message boundaries in data. Rdm (Reliably Delivered Messages) messages arrive unduplicated and in order. Furthermore, the sender is notified if messages are lost. If you initialize a Socket using Rdm, you do not require a remote host connection before sending and receiving data. With Rdm, you can communicate with multiple peers.'),
        5 : ('Seqpacket', 'Provides connection-oriented and reliable two-way transfer of ordered byte streams across a network. Seqpacket does not duplicate data, and it preserves boundaries within the data stream. A Socket of type Seqpacket communicates with a single peer and requires a remote host connection before communication can begin.'),
        1 : ('Stream', 'Supports reliable, two-way, connection-based byte streams without the duplication of data and without preservation of boundaries. A Socket of this type communicates with a single peer and requires a remote host connection before communication can begin. Stream uses the Transmission Control Protocol (ProtocolType.Tcp) and the AddressFamily.InterNetwork address family.'),
        -1 : ('Unknown', 'Specifies an unknown Socket type.'),
    })

    def __init__(self, *, c: Color | None = None, parent: Optional[Vertex] = None) -> None:
        super().__init__(c=c, parent=parent)
        self.family : str = ""
        self.protocol : str = ""
        self.type : str = ""





class Host(UniqueVertex):

    """
    A machine vertex. It represents a physical machine.
    """

    __current : Optional["Host"] = None

    @staticproperty
    def current() -> "Host":
        """
        The currently active Host node. This is used to indicate which Host node is the machine running the sample.
        """
        if Host.__current is None:
            raise AttributeError("Host class has no attribute 'current'.")
        return Host.__current

    @current.setter
    def current(value : "Host"):
        if not isinstance(value, Host):
            raise TypeError("Cannot set attribute 'current' of class 'Host' to object of type '{}'".format(type(value).__name__))
        Host.__current = value

    __slots__ = {
        "address" : "The IP address of the machine",
        "domain" : "The URL the machine is known as if any",
        "name" : "The machine's name",
        "platform" : "The operating system the host is running on"
    }

    __pickle_slots__ = {
        "address",
        "domain",
        "name",
        "platform"
    }

    def __init__(self, *, c: Color | None = None, parent: Optional[Vertex] = None) -> None:
        super().__init__(c=c, parent=parent)
        self.address : str = ""
        self.domain : str = ""
        self.name : str = ""
        self.platform : str = "Unknown"
    
    @property
    def label(self) -> str:
        """
        Returns a label for this Host node.
        """
        if self.name:
            return "Host " + repr(self.name)
        return "Host at " + self.address





class SpawnedProcess(Edge):
    
    """
    This kind of edge indicates that a machine hosts a process.
    """

    source : Host
    destination : Process





class HasSocket(Edge):

    """
    This kind of edge indicates that a process opened a socket.
    """

    label : str = ""

    source : Process
    destination : Socket





class HasConnection(Edge):

    """
    This kind of edge indicates that a socket makes a connection.
    """

    label : str = ""

    source : Socket
    destination : Connection





class Communicates(Edge):

    """
    This kind of edge indicates that two hosts communicate via a connection
    """

    source : Connection
    destination : Host





class CreatesSocket(Edge):

    """
    This kind of edge indicates that a system call created a socket.
    """

    label : str = ""

    source : Call
    destination : Socket





class Binds(Edge):

    """
    This kind of edge indicates that a system call bound a connection to a local address.
    """

    __slots__ = {
        "src" : "The source (local) address of the connection",
    }

    __pickle_slots__ = {
        "src"
    }

    label : str = ""

    source : Call
    destination : Connection

    def __init__(self, source: Vertex, destination: Vertex, *, c: Color | None = None, auto_write: bool = True) -> None:
        super().__init__(source, destination, c=c, auto_write=auto_write)
        self.src : Any = None





class Connects(Edge):

    """
    This kind of edge indicates that a system call connected a connection to a remote address.
    """

    __slots__ = {
        "dest" : "The destination (remote) address of the connection"
    }

    __pickle_slots__ = {
        "dest"
    }

    label : str = ""

    source : Call
    destination : Connection

    def __init__(self, source: Vertex, destination: Vertex, *, c: Color | None = None, auto_write: bool = True) -> None:
        super().__init__(source, destination, c=c, auto_write=auto_write)
        self.dest : Any = None





class Sends(Arrow):

    """
    This kind of arrow indicates that a system call sent data through a connection.
    """

    from typing import TYPE_CHECKING
    if TYPE_CHECKING:
        from .data import Data
    del TYPE_CHECKING

    source : Call
    destination : "Data"





class Receives(Arrow):

    """
    This kind of arrow indicates that a system call received data through a connection.
    """

    from typing import TYPE_CHECKING
    if TYPE_CHECKING:
        from .data import Data
    del TYPE_CHECKING

    source : "Data"
    destination : Call





class Conveys(Arrow):
    
    """
    This kind of arrow indicates that a connection conveyed a message.
    """

    label : str = ""

    from typing import TYPE_CHECKING
    if TYPE_CHECKING:
        from .data import Data
    del TYPE_CHECKING

    source : "Connection | Data"
    destination : "Connection | Data"





class Closes(Edge):

    """
    This kind of edge indicates that a system call closed a connection.
    """
    
    label : str = ""

    source : Call
    destination : Connection





class CloseSocket(Edge):

    """
    This kind of edge indicates that a system call closed a socket object.
    """

    label : str = ""

    source : Call
    destination : Socket





class ListensOn(Edge):

    """
    This kind of edge indicates that a system call set a socket to listening mode.
    """

    source : Call
    destination : Socket





class Shutdown(Edge):

    """
    This kind of edge indicates that a connection was shutdown by a system call.
    """

    label : str = ""

    source : Call
    destination : Connection





class Accepts(Edge):

    """
    This kind of arrow indicates that a system call accepted a connection from a remote address.
    """

    __slots__ = {
        "dest" : "The destination (remote) address of the connection"
    }

    __pickle_slots__ = {
        "dest"
    }

    source : Call
    destination : Connection

    def __init__(self, source: Vertex, destination: Vertex, *, c: Color | None = None, auto_write: bool = True) -> None:
        super().__init__(source, destination, c=c, auto_write=auto_write)
        self.dest : Any = None





class Duplicates(Arrow):

    """
    This kind of arrow indicates that a socket was duplicated to form a similar socket (example: after a call to accept()).
    """

    source : Socket
    destination : Socket





_active_connections : dict[Socket, Connection] = {}
_active_handles : dict[Socket, int] = {}
_inverted_handles : dict[int, Socket] = {}

@chrono
def integrate_socket_creation(c : Call):
    """
    Creates a Socket vertex and links it to the call that created it.
    """
    from ...logger import logger
    if c.status == 1:
        logger.debug("Socket created.")
        s = Socket()
        _active_handles[s] = c.arguments.socket
        _inverted_handles[c.arguments.socket] = s
        s.family = Socket.families[c.arguments.af][0]
        s.protocol = Socket.protocols[c.arguments.protocol][0]
        s.type = Socket.types[c.arguments.type][0]
        if s in _active_connections:
            _active_connections.pop(s)
        CreatesSocket(c, s)
        HasSocket(c.thread.process, s)

@chrono
def integrate_socket_binding(c : Call):
    """
    Creates a Connection vertex and binds it to a local address.
    """
    from ...logger import logger
    from ..graph import find_or_create
    if c.status == 1:
        logger.debug("Socket bound.")
        l = Connection()
        if c.arguments.socket in _inverted_handles:
            s = _inverted_handles[c.arguments.socket]
        else:
            logger.warning("Binding to non-existing socket.")
            return
        _active_connections[s] = l
        HasConnection(s, l)
        Communicates(l, find_or_create(Host, domain = "host")[0])
        b = Binds(c, l)
        if s.family == "InterNetwork":
            b.src = (c.arguments.ip_address, c.arguments.port)
            l.src = b.src
        else:
            logger.error("I don't know how to handle this type of address:\n{}".format(s))

@chrono
def integrate_socket_connection(c : Call):
    """
    Sets the remote address of a connection.
    """
    from ...logger import logger
    from ..graph import find_or_create
    if c.status == 1:
        logger.debug("Socket connected.")
        try:
            sock = c.arguments.socket
        except AttributeError:
            sock = c.arguments.s
        if sock in _inverted_handles:
            s = _inverted_handles[sock]
        else:
            logger.warning("Connecting from non-existing socket.")
            return
        if s in _active_connections:
            l = _active_connections[s]
            logger.warning("Trying to connect an unbound socket...")
        else:
            l = Connection()
        HasConnection(s, l)
        Communicates(l, find_or_create(Host, domain = "host")[0])
        _active_connections[s] = l
        b = Connects(c, l)
        Communicates(l, find_or_create(Host, address = c.arguments.ip_address)[0])
        if s.family == "InterNetwork":
            b.dest = (c.arguments.ip_address, c.arguments.port)
            l.dest = b.dest
        else:
            logger.error("I don't know how to handle this type of address:\n{}".format(s))


@chrono
def integrate_socket_listening(c : Call):
    """
    Sets the socket in listening mode.
    """
    from ...logger import logger
    if c.status == 1:
        logger.debug("Socket listening.")
        if c.arguments.socket in _inverted_handles:
            s = _inverted_handles[c.arguments.socket]
        else:
            logger.warning("Listening from non-existing socket.")
            return
        ListensOn(c, s)

@chrono
def integrate_socket_accepting(c : Call):
    """
    Sets a connection remote address through an accept system call. Also creates a new socket object.
    """
    from ...logger import logger
    from ..graph import find_or_create
    if c.status == 1:
        logger.debug("Socket accepting connection.")
        if c.arguments.socket in _inverted_handles:
            s = _inverted_handles[c.arguments.socket]
        else:
            logger.warning("Accepting from non-existing socket.")
            return
        s1 = Socket()
        l1 = Connection()
        s1.family = s.family
        s1.protocol = s.protocol
        s1.type = s.type
        _active_handles[s1] = c.return_value
        _inverted_handles[c.return_value] = s1
        if s not in _active_connections:
            logger.warning("Accepting from unbound socket...")
        else:
            l = _active_connections[s]
            l1.src = l.src
        b = Accepts(c, l1)
        HasConnection(s1, l1)
        Duplicates(s, s1)
        Communicates(l1, find_or_create(Host, address = c.arguments.ip_address)[0])
        if s1.family == "InterNetwork":
            b.dest = (c.arguments.ip_address, c.arguments.port)
            l1.dest = b.dest

@chrono
def integrate_socket_close(c : Call):
    """
    Closes a socket and the connection linked to it.
    """
    from ...logger import logger
    if c.status == 1:
        logger.debug("Closing socket.")
        if c.arguments.socket in _inverted_handles:
            s = _inverted_handles[c.arguments.socket]
        else:
            logger.warning("Closing non-existing socket.")
            return
        CloseSocket(c, s)
        handle = _active_handles.pop(s)
        _inverted_handles.pop(handle)
        if s in _active_connections:
            l = _active_connections.pop(s)
            Closes(c, l)
        else:
            logger.warning("Closing down unbound socket...")

@chrono
def integrate_socket_shutdown(c : Call):
    """
    Closes a connection.
    """
    from ...logger import logger
    if c.status == 1:
        logger.debug("Closing socket.")
        if c.arguments.socket in _inverted_handles:
            s = _inverted_handles[c.arguments.socket]
        else:
            logger.warning("Shuting down non-existing socket")
            return
        if s in _active_connections:
            l = _active_connections.pop(s)
            Shutdown(c, l)
        else:
            logger.warning("Shuting down unbound socket...")

@chrono
def integrate_socket_send(c : Call):
    """
    Sends data through a connection.
    """
    from ...logger import logger
    from .data import Data, register_write_operation
    if c.status == 1:
        logger.debug("Sending data.")
        if c.arguments.socket in _inverted_handles:
            s = _inverted_handles[c.arguments.socket]
        else:
            logger.warning("Sending through non-existing socket.")
            return
        if s not in _active_connections:
            logger.warning("Sending through unbound socket.")
        else:
            l = _active_connections[s]
            d = Data()
            d.time = c.time
            d.set_data(c.arguments.buffer)
            Sends(c, d)
            Conveys(d, l)
            l.volume += len(d.data)
            register_write_operation(l, s, c.arguments.buffer)

@chrono
def integrate_socket_recv(c : Call):
    """
    Receives data through a connection.
    """
    from ...logger import logger
    from .data import Data, register_read_operation
    if c.status == 1:
        logger.debug("Receiving data.")
        if c.arguments.socket in _inverted_handles:
            s = _inverted_handles[c.arguments.socket]
        else:
            logger.warning("Receiving through non-existing socket.")
            return
        if s not in _active_connections:
            logger.warning("Receiving through unbound socket.")
        else:
            l = _active_connections[s]
            d = Data()
            d.time = c.time
            d.set_data(c.arguments.buffer)
            Receives(d, c)
            Conveys(l, d)
            l.volume += len(d.data)
            register_read_operation(l, s, c.arguments.buffer)





# Socket creation
CallHandler(lambda c : c.name in ("socket", "WSASocketW"), integrate_socket_creation)

# Socket binding
CallHandler(lambda c : c.name == "bind", integrate_socket_binding)

# Socket connecting
CallHandler(lambda c : c.name in ("connect", "WSAConnect"), integrate_socket_connection)

# Socket listening
CallHandler(lambda c : c.name == "listen", integrate_socket_listening)

# Socket accepting
CallHandler(lambda c : c.name == "accept", integrate_socket_accepting)

# Socket closing
CallHandler(lambda c : c.name == "closesocket", integrate_socket_close)

# Socket shuting down
CallHandler(lambda c : c.name == "shutdown", integrate_socket_shutdown)

# Socket sending
CallHandler(lambda c : c.name == "send", integrate_socket_send)

# Socket receiving
CallHandler(lambda c : c.name == "recv", integrate_socket_recv)


del defaultdict, logger, Any, Optional, Color, Arrow, UniqueVertex, Vertex, Call, CallHandler, chrono, integrate_socket_accepting, integrate_socket_binding, integrate_socket_close, integrate_socket_connection, integrate_socket_creation, integrate_socket_listening, integrate_socket_recv, integrate_socket_send, integrate_socket_shutdown


import json
import logging
import os
from threading import RLock, Thread, Event
from time import sleep
from code import interact
from .source.build import *
from .source.database import *
from .source.utils import extract
import argparse

parser = argparse.ArgumentParser(description="Parses all Cuckoo reports in given paths, classify their API calls, and stores them in Python databases.")

parser.add_argument("-r", "--recursive", action="store_true", help="If true, will explore given folders recursively to find Cuckoo reports.")

args1, remaining_args = parser.parse_known_args()

work : set[str] = set()
lock = RLock()

def extend_work(s : str):
    from pathlib import Path
    import glob
    with lock:
        for s in glob.iglob(s):
            p = Path(s).absolute()
            if p.is_dir():
                for pi in p.iterdir():
                    if pi.is_dir() and args1.recursive:
                        extend_work(pi)
                    elif pi.is_file() and pi.suffix == ".json":
                        work.add(pi)
            elif p.is_file() and args1.recursive:
                work.add(p)

parser.add_argument("sources", type = extend_work, action="extend", nargs = "+", help="The source files/folders to parse for API calls.")
parser.add_argument("--all_categories", action="store_true", help="If enabled, won't start an interactive prompt, but instead will compile one database for all available category (not name) of API call")

args = parser.parse_args(remaining_args)

DATA_PATH = "database.pyt"

database = Database(dict, List[List[List[List[dict]]]])
# database = Database(dict, List[dict])
running = Event()
finished = Event()


def start():
    with lock:
        running.set()
        finished.clear()

def stop():
    running.clear()

def save(path = None):
    from pickle import dump
    if path == None:
        path = DATA_PATH
    with lock:
        with open(path, "wb") as f:
            dump(database, f)
        print("Succesfully saved database.")

def size():
    with lock:
        print("Got {} types of calls and {} calls in total.".format(len(database), sum(len(li) for li in database.values())))

def remaining():
    return len(work)

def secured(func, *args, **kwargs):
    with lock:
        return func(*args, **kwargs)

fil = lambda x : True

def explorer():
    global fil
    while True:
        running.wait()
        with lock:
            if work:
                file = work.pop()
            else:
                finished.set()
                sleep(0.01)
                continue
        try:
            with open(file, "rb") as f:
                data = json.load(f)
        except:
            # print("Exception while loading file '{}':".format(file))
            # print_exc()
            continue
        E : List[List[List[dict]]] = []
        database.database.append(E)
        try:
            for p in data["behavior"]["processes"]:
                P : List[List[dict]] = []
                table : Dict[int, list] = {}
                for c in p["calls"]:
                    with lock:
                        running.wait()
                        try:
                            if fil(c):
                                if c["tid"] not in table:
                                    table[c["tid"]] = []
                                    P.append(table[c["tid"]])
                                table[c["tid"]].append(c)
                        except:
                            print(c, len(database))
                            raise
                E.append(P)
        except:
            print("Exception while analyzing file {}:".format(file))
        # for p in bf.data["behavior"]["processes"]:
        #     for c in p["calls"]:
        #         with lock:
        #             running.wait()
        #             try:
        #                 database.database.append(c)
        #             except:
        #                 print(c, len(database))
        #                 raise

def set_filter(f : Callable[[dict], bool]):
    global fil
    fil = f

def at_end(f, *args, **kwargs):
    def _wait():
        while remaining():
            sleep(0.001)
        f(*args, **kwargs)
    Thread(target = _wait, daemon = True).start()

categories = (
    "registry",
    "__notification__",
    "certificate",
    "crypto",
    "exception",
    "file",
    "iexplore",
    "misc",
    "netapi",
    "network",
    "ole",
    "process",
    "resource",
    "services",
    "synchronisation",
    "system",
    "ui"
)


t = Thread(target = explorer, daemon = True)
t.start()


if not args.all_categories:

    from Viper.interactive import InteractiveInterpreter

    env = {
        "start" : start,
        "stop" : stop,
        "save" : save,
        "size" : size,
        "remaining" : remaining,
        "search" : extend_work,
        "secured" : secured,
        "database" : database,
        "extract" : extract,
        "set_filter" : set_filter,
        "at_end" : at_end, 
        "logging" : logging,
        "categories" : categories,
        "finished" : finished
    }

    InteractiveInterpreter(env).interact("Interactive API call archiver. Use dir() to see all the available functions.")

else:

    initial_work = work.copy()

    print("About to compute databases for {} categories, using {} Cuckoo reports.".format(len(categories), len(initial_work)))

    for cat in categories:
        if not os.path.exists(cat + ".pyt"):
            database.clear()
            work.clear()
            work |= initial_work
            set_filter(lambda c : c["category"] == cat)
            print("Compiling database for category '{}'.".format(cat))
            start()
            while not finished.wait(20):
                print("Progress : {:.2f}%".format((len(initial_work) - remaining()) / remaining() * 100))
            stop()

            print("Creating 'names' view.")
            database.create_view("names", extract("api"))

            print("Saving database '{}'.".format(cat + ".pyt"))
            save(cat + ".pyt")
        else:
            print(cat + ".pyt database already exists.")


from types import UnionType




class MetaMatch:

    """
    This class encapsulates all the information about a given metagraph match in a specific graph.
    """

    __slots__ = {
        "file" : "The Python graph file that the match was made in.",
        "meta" : "The name of the metagraph that matched.",
        "graph" : "The subgraph that matched the metagraph." 
    }

    def __init__(self, file, meta, graph):
        self.file = file
        self.meta = meta
        self.graph = graph





def class_union(*cls : type) -> type | UnionType:
    """
    Creates the union of classes, as sum operator but for the "|" operator.
    """
    for c in cls:
        if not isinstance(c, type):
            raise TypeError("Expected types, got " + repr(type(c).__name__))
    l = list(cls)
    if not l:
        raise ValueError("Expected at least one type.")
    c = l.pop()
    while l:
        c = c | l.pop()
    return c


"""
This is the MetaLib! It is a library for MetaGraphs, a pattern storage.

All existing MetaGraphs are loaded with the package.
To create new MetaGraphs, use the module croutons.metalib.interactive.
"""

from ...bakery.logger import logger
from .utils import entries, load

logger.info("Loading MetaGraph library.")

__all__ = []





g = globals()
for name in entries():
    g[name] = load(name)
    __all__.append(name)
    del name





del logger, load, entries, g


"""
This module contains configuration objects for execution graphs.
"""

from typing import Any, Type
from .colors import Color
from .graph import Vertex, Edge
from Viper.meta.decorators import hybridmethod





class ColorMap:

    """
    Just a class to hold all basic colors for the graph.
    Create instances to customize!
    """

    __current = None

    default = Color.white

    Deleted = Color(100, 100, 100)

    class execution:
        Process = Color(255, 255, 50)
        Thread = Color(255, 204, 0)
        Call = Color(255, 153, 0)
        HasChildProcess = Color.black
        HasThread = Color.black

    class network:
        Host = Color(255, 255, 255)
        Connection = Color(100, 50, 255)
        Socket = Color.white
        HasProcess = Color.black
        SpawnedProcess = Color.black
    
    network.Socket = Color.average(network.Connection, execution.Process)

    class filesystem:
        File = Color(0, 255, 50)
        Directory = Color(0, 100, 0)
        Handle = Color.white
    
    filesystem.Handle = Color.average(filesystem.File, execution.Process)

    class data:
        Data = Color(0, 255, 150) + Color(150, 0, 255)
        Diff = Color(50, 150, 255)
        DiffLowEntropy = Color(50, 150, 255)
        DiffHighEntropy = Color(255, 50, 150)
    
    class imports:
        Import = Color(150, 150, 0)
        SuspiciousImport = Color(255, 150, 0)
    
    class registry:
        Key = Color(0, 150, 255)
        UnprintableKey = Color(255, 0, 50)
        KeyEntry = Color(0, 255, 255)
        Handle = Color.white

    registry.Handle = Color.average(registry.Key, execution.Process)

    @hybridmethod
    def get_color(self, cls : Type[Vertex] | Type[Edge]) -> Color:
        """
        This function returns the color corresponding to the given class of Vertex or Edge.
        """
        from inspect import getmodule
        try:
            return getattr(getattr(self, getmodule(cls).__name__.split(".")[-1]), cls.__name__)
        except AttributeError:
            return self.default
    
    @hybridmethod
    def has_color(self, cls : Type[Vertex] | Type[Edge]) -> bool:
        """
        This function returns True if the given class of Vertex or Edge has a special color.
        """
        from inspect import getmodule
        try:
            getattr(getattr(self, getmodule(cls).__name__.split(".")[-1]), cls.__name__)
            return True
        except AttributeError:
            return False
        
    @staticmethod
    def get_config() -> Type["ColorMap"] | "ColorMap":
        """
        Returns the globally set ColorMap configuration object.
        """
        return ColorMap.__current
    
    @hybridmethod
    def set_config(self):
        """
        Sets the calling object (class or instance) as the globally used configuration for ColorMap.
        """
        ColorMap.__current = self

ColorMap.set_config()





class SizeMap:

    """
    Just a class to hold all basic sizes for the graph.
    Create instances to customize!
    """

    __current = None

    default = 2.0

    class execution:
        Process = 5.0
        Thread = 2.0
        Call = .3
    
    class network:
        Host = 10.0
        Socket = 2.5
        Connection = 1.5

    class filesystem:
        File = 2.5
        Directory = 2.5
        Handle = 1.5
    
    class data:
        Data = .5
        Diff = 1.5

    class imports:
        Import = 0.75
    
    class registry:
        Key = 1.5
        KeyEntry = 1.0
        Handle = 1.0

    @hybridmethod
    def get_size(self, cls : Type[Vertex]) -> float:
        """
        This function returns the size corresponding to the given class of Vertex.
        """
        from inspect import getmodule
        try:
            return getattr(getattr(self, getmodule(cls).__name__.split(".")[-1]), cls.__name__)
        except AttributeError:
            return self.default
    
    @hybridmethod
    def has_size(self, cls : Type[Vertex]) -> bool:
        """
        This function returns True if the given class of Vertex has a special size.
        """
        from inspect import getmodule
        try:
            getattr(getattr(self, getmodule(cls).__name__.split(".")[-1]), cls.__name__)
            return True
        except AttributeError:
            return False
        
    @staticmethod
    def get_config() -> Type["SizeMap"] | "SizeMap":
        """
        Returns the globally set SizeMap configuration object.
        """
        return SizeMap.__current
    
    @hybridmethod
    def set_config(self):
        """
        Sets the calling object (class or instance) as the globally used configuration for SizeMap.
        """
        SizeMap.__current = self

SizeMap.set_config()





class CompilationParameters:

    """
    This class describes some compilations parameters that some modules may want to use.
    """

    __current = None

    SkipLevenshteinForDataNodes = False
    SkipLevenshteinForDiffNodes = False

    @hybridmethod
    def get_param(self, name : str) -> Any:
        """
        This function returns the parameter corresponding to the given name.
        Raises AttributeError if the parameter does not exist.
        """
        try:
            return getattr(self, name)
        except AttributeError:
            raise AttributeError("No such parameter : " + repr(name))
    
    @hybridmethod
    def has_color(self, name : str) -> bool:
        """
        This function returns True if the given parameter exists.
        """
        try:
            getattr(self, name)
            return True
        except AttributeError:
            return False
    
    @staticmethod
    def get_config() -> Type["CompilationParameters"] | "CompilationParameters":
        """
        Returns the globally set CompilationParameters configuration object.
        """
        return CompilationParameters.__current
    
    @hybridmethod
    def set_config(self):
        """
        Sets the calling object (class or instance) as the globally used configuration for CompilationParameters.
        """
        CompilationParameters.__current = self

CompilationParameters.set_config()


from numbers import Complex
from typing import Any, Callable, Generic, Iterable, Iterator, ParamSpec, Sequence, Set, Type, TypeVar
import pathlib


from Viper.debugging.chrono import Chrono
chrono = Chrono()
chrono.enabled = False
del Chrono





def avg(*s : Complex | Sequence[Complex], **kwargs : Any) -> Complex:
    """
    Returns the average of a series of numbers. Works like min/max without the key argument.
    """
    for k in kwargs:
        if k != "default":
            raise TypeError(repr(k) + " is an invalid keyword argument for avg()")
    if len(s) == 0:
        raise TypeError("avg expected at least 1 argument, got 0")
    if len(s) == 1:
        s = s[0]
    n, m = 0, 0
    for e in s:
        n += 1
        m += e
    if n == 0:
        if "default" in kwargs:
            return kwargs["default"]
        raise ZeroDivisionError("avg() arg is an empty sequence")
    return n / m


def dev(*s : Complex | Sequence[Complex], **kwargs : Any) -> Complex:
    """
    Returns the standart deviation of a series of numbers. Works like min/max without the key argument.
    """
    for k in kwargs:
        if k != "default":
            raise TypeError(repr(k) + " is an invalid keyword argument for dev()")
    s = list(s)
    if len(s) == 0:
        if "default" in kwargs:
            return kwargs["default"]
        raise ZeroDivisionError("dev() arg is an empty sequence")
    a = avg(s)
    return avg((x - a) ** 2 for x in s) ** 0.5


class extract:

    def __init__(self, *keys : Any) -> None:
        self.keys = keys
    
    def __call__(self, c : dict) -> Any:
        for k in self.keys:
            c = c[k]
        return c


def pretty_print(c : list | dict):
    """
    JSON style pretty print.
    """
    import json
    print(json.dumps(c, indent="\t"))


def extract_pickle_slots(o : type) -> Set[str]:
    if not hasattr(o, "__pickle_slots__"):
        s = set()
    else:
        s = set(o.__pickle_slots__)
    return s.union(*[extract_pickle_slots(b) for b in o.__bases__])


def path_factory(path : str) -> pathlib.PurePath:
    """
    Returns the correct path factory class for the given graph (Either WindowsPurePath or PosixPurePath).
    """
    from .graph import Graph
    from pathlib import PureWindowsPath, PurePosixPath
    platform = Graph.active_graphs()[-1].data["platform"].lower() if Graph.active_graphs() else "windows"
    if "windows" in platform:
        return PureWindowsPath(path)
    else:
        return PurePosixPath(path)


def is_path(s : str) -> bool:
    """
    Returns True if the given string represents a path for the current platform.
    """
    from .graph import Graph
    from pathlib import PureWindowsPath, PurePosixPath
    platform = Graph.active_graphs()[-1].data["platform"].lower() if Graph.active_graphs() else "windows"
    if "windows" in platform:
        forbidden = set("<>\"|?*")
        if forbidden.intersection(s):           # Contains characters that are forbidden in Windows paths.
            return False
        if s.startswith("-") or s.startswith("/"):  # Could be a path, but it is more likely to be flag
            return False
        try:
            if PureWindowsPath(s).is_reserved():
                return False
            return True
        except:
            return False
    else:
        if s.startswith("-"):                       # Could be a path, but it is more likely to be flag
            return False
        try:
            if PurePosixPath(s).is_reserved():
                return False
            return True
        except:
            return False
        
def parse_command_line(s : str) -> list[str]:
    """
    Parses command line with the syntax of the current platform's shell. Returns the argument vector.
    """
    import re
    from .graph import Graph

    platform = Graph.active_graphs()[-1].data["platform"].lower() if Graph.active_graphs() else "windows"
    if "windows" not in platform:
        RE_CMD_LEX = r'''"((?:\\["\\]|[^"])*)"|'([^']*)'|(\\.)|(&&?|\|\|?|\d?\>|[<])|([^\s'"\\&|<>]+)|(\s+)|(.)'''
    else:
        RE_CMD_LEX = r'''"((?:""|\\["\\]|[^"])*)"?()|(\\\\(?=\\*")|\\")|(&&?|\|\|?|\d?>|[<])|([^\s"&|<>]+)|(\s+)|(.)'''

    args : list[str] = []
    accu = None   # collects pieces of one arg
    for qs, qss, esc, pipe, word, white, fail in re.findall(RE_CMD_LEX, s):
        if word:
            pass   # most frequent
        elif esc:
            word = esc[1]
        elif white or pipe:
            if accu is not None:
                args.append(accu)
            if pipe:
                args.append(pipe)
            accu = None
            continue
        elif fail:
            raise ValueError("invalid or incomplete shell string")
        elif qs:
            word = qs.replace('\\"', '"').replace('\\\\', '\\')
            if "windows" in platform:
                word = word.replace('""', '"')
        else:
            word = qss   # may be even empty; must be last

        accu = (accu or '') + word

    if accu is not None:
        args.append(accu)

    return args


P = TypeVar("P")
P2 = TypeVar("P2")
CB_Args = ParamSpec("CB_Args")

class System(Generic[P]):

    def __init__(self) -> None:
        from typing import Callable
        self.__system : set[P] = set()
        self.__population : set[P] = set()
        self.__at_collapse : list[tuple[Callable[..., None], tuple, dict[str, Any]]] = []
        collapse = self.__at_collapse
        population = self.__population
        system = self.__system

        class Particule(Generic[P2]):

            __slots__ = {"__particule"}

            def __init__(self, particule : P2) -> None:
                self.__particule = particule

            def __enter__(self):
                if self.__particule in system:
                    raise RuntimeError("Trying to entangle Particule with itself : cannot copy quantum information.")
                population.discard(self.__particule)    # type: ignore
                system.add(self.__particule)            # type: ignore
           
            def __exit__(self, exc_type, exc_value, traceback):
                if self.__particule not in system:
                    raise RuntimeError("Trying to collapse non-quantum object.")
                system.remove(self.__particule)         # type: ignore
                if not system:  # The system has fully collapsed : time for callbacks
                    for cb, args, kwargs in collapse:
                        cb(*args, **kwargs)
                    collapse.clear()
        
        self.Entangled : Type[Particule[P]] = Particule

    def __contains__(self, p : P) -> bool:
        return p in self.__system
    
    def __len__(self) -> int:
        return len(self.__system)
    
    @property
    def population(self) -> set[P]:
        """
        A set of elements that can be included in the system.
        """
        return self.__population
    
    @population.setter
    def population(self, v : Iterable[P]):
        from typing import Iterable
        if not isinstance(v, Iterable):
            raise TypeError("Expected iterable, got " + repr(type(v).__name__))
        self.__population.clear()
        self.__population.update(k for k in v if k not in self)

    def include(self) -> Iterator[P]:
        """
        Yields all successives items of the population to be added to the system.
        """
        while self.__population:
            yield self.__population.pop()
    
    def add_callback(self, cb : Callable[CB_Args, None], *args : CB_Args.args, **kwargs : CB_Args.kwargs):
        """
        Adds a callable to be called with given arguments when the system completely resolves (i.e. when all entangled wave functions have collapsed).
        """
        if not self.__system:
            raise RuntimeError("Cannot program a callback before the experiment has started. Add at least one particle first.")
        self.__at_collapse.append((cb, args, kwargs))
    
    def remove_callback(self, cb : Callable[CB_Args, None]):
        """
        Removes all callbacks to the given function.
        """
        to_pop = []
        for i, (cbi, argsi, kwargsi) in enumerate(self.__at_collapse):
            if cb == cbi:
                to_pop.append(i)
        for i in reversed(to_pop):
            self.__at_collapse.pop(i)


"""
The event library. An event can be created by anyone, specifying any kind of data. Anybody can wait for any type of events.
"""

from typing import Callable, Generic, Type, TypeVar
from Viper.better_threading import Future
from threading import Lock
from .utils import chrono

__all__ = ["Event"]





T = TypeVar("T", bound="Event")

class Event(Generic[T]):

    """
    The base event class. Use its static methods to react to Events created.
    """

    __waiting : dict[Type[T], list[Future[T]]] = {}
    __callbacks : dict[Type[T], list[Callable[[T], None]]] = {}
    __locks : dict[Type[T], Lock] = {}

    __slots__ = {}

    def __str__(self) -> str:
        """
        Implements str(self).
        """
        return type(self).__name__ + "(" + ", ".join(str(name) + " = " + repr(getattr(self, name)) for name in self.__slots__) + ")"

    @chrono
    def throw(self : T) -> bool:
        """
        Triggers the Event. Everyone waiting for it will be awaken, and corresponding callbacks will be called.
        Returns True if anybody reacted to the event. Returns False otherwise.
        """
        from ..logger import logger

        def raise_multiple_exceptions(exceptions : list[BaseException]):
            """
            Raises all the exceptions in the given list, in order.
            """
            if not exceptions:
                return
            try:
                raise exceptions.pop() from None
            finally:
                raise_multiple_exceptions(exceptions)

        classes = {type(self)}
        done = set()
        ok = False
        logger.debug("Throwing {}.".format(self))
        while classes:
            cls = classes.pop()

            with Event.__locks[cls]:

                for fut in Event.__waiting[cls]:
                    fut.set(self)
                    ok = True
                Event.__waiting[cls].clear()
                
                exceptions : list[BaseException] = []
                for cb in Event.__callbacks[cls]:
                    try:
                        cb(self)
                        ok = True
                    except BaseException as e:
                        exceptions.append(e)
                raise_multiple_exceptions(exceptions)

            done.add(cls)
            classes.update(cls.__subclasses__())
            classes.difference_update(done)
        
        return ok

    @classmethod
    def __init_subclass__(cls : Type[T]) -> None:
        from threading import Lock
        Event.__locks[cls] = Lock()
        Event.__callbacks[cls] = []
        Event.__waiting[cls] = []

    @classmethod
    def wait(cls : Type[T]) -> T:
        """
        Waits for an Event of this class or of any subclasses to occur. Returns the Event object when it happens.
        """
        from Viper.better_threading import Future
        fut : Future[T] = Future()
        with Event.__locks[cls]:
            Event.__waiting[cls].append(fut)
        return fut.result()
    
    @classmethod
    def add_callback(cls : Type[T], callback : Callable[[T], None]):
        """
        Adds a callback to be performed after a realization of an event of this class or of any of its subclasses. The callback will persist until it is removed.
        """
        with Event.__locks[cls]:
            Event.__callbacks[cls].append(callback)
    
    @classmethod
    def remove_callback(cls : Type[T], callback : Callable[[T], None]) -> bool:
        """
        Removes a callback to be performed after a realization of an event with the given name. Returns True if a callback was indeed deleted.
        """
        with Event.__locks[cls]:
            try:
                Event.__callbacks[cls].remove(callback)
                return True
            except ValueError:
                return False


Event.__init_subclass__()       # It must initialize itself, as if it was one of its own subclasses.

del Callable, Type, Future, Lock


"""
This module defines basic graph filters. Use them to reduce the size of the graph by only keeping specific edges.
To be used with compile.py, all custom Filters must be defined in this file!
"""

from typing import Callable, Iterable
from .graph import Edge, Vertex, Graph
from Viper.meta.iterable import InstanceReferencingClass
from .utils import chrono, System
from .types.registry import Key, KeyEntry, Handle
from functools import cache

__all__ = ["Filter", "significant_call_only", "no_data_nodes", "no_simple_imports", "no_handle_nodes", "injected_threads_only", "significant_processes_only", "modified_registry_only"]





class Filter(metaclass = InstanceReferencingClass):

    """
    A filter object for graphs.
    """

    def __init__(self, edge_filter_function : Callable[[Edge], bool], vertex_filter_function : Callable[[Vertex], bool]) -> None:
        self.__edge_filter = edge_filter_function
        self.__vertex_filter = vertex_filter_function
    
    def test_vertex(self, u : Vertex) -> bool:
        """
        Tests if a given vertex should be kept in the graph.
        """
        from .graph import Vertex
        if not isinstance(u, Vertex):
            raise TypeError("Expected Vertex, got " + repr(type(u).__name__))
        return self.__vertex_filter(u)
    
    def test_edge(self, e : Edge) -> bool:
        """
        Tests if a given edge should be kept in the graph.
        """
        from .graph import Edge
        if not isinstance(e, Edge):
            raise TypeError("Expected Edge, got " + repr(type(e).__name__))
        return self.__edge_filter(e)
    
    def __call__(self, iterable : Iterable[Edge | Vertex]) -> Iterable[Edge | Vertex]:
        """
        Implements self(iterable). Filters the given Egde | Vertex iterable.
        """
        from .graph import Edge, Vertex
        from typing import Iterable
        if not isinstance(iterable, Iterable):
            raise TypeError("Expected iterable, got " + repr(type(iterable).__name__))
        for x in iterable:
            if isinstance(x, Edge):
                if self.test_edge(x):
                    yield x
            elif isinstance(x, Vertex):
                if self.test_vertex(x):
                    yield x
            else:
                raise TypeError("Expected iterable of Edges or Vertices, got a " + repr(type(x).__name__))
    
    @chrono
    def apply(self, g : Graph):
        """
        Applies the filter to the given graph, modifying it.
        """
        from .graph import Graph
        if not isinstance(g, Graph):
            raise TypeError("Expected Graph, got " + repr(type(g).__name__))
        edges_to_remove = {e for e in g.edges if not self.__edge_filter(e)}
        vertices_to_remove = {u for u in g.vertices if not self.__vertex_filter(u)}
        for e in edges_to_remove:
            g.remove(e)
        for u in vertices_to_remove:
            g.remove(u)





# Here we will define some basic filters:

def important_call_filter(u : Vertex) -> bool:
    from .types.execution import Call
    if not isinstance(u, Call):
        return True
    else:
        for v in u.neighbors():
            if not isinstance(v, Call):
                return True
        return False

def data_filter(u : Vertex) -> bool:
    from .types.data import Data
    return not isinstance(u, Data)

def simple_imports(u : Vertex) -> bool:
    from .types.imports import Import
    from .types.execution import Process
    return not isinstance(u, Import) or bool([v for v in u.neighbors() if not isinstance(v, Process)])

def no_handles(u : Vertex) -> bool:
    from .types.filesystem import Handle as Handle
    from .types.network import Socket
    from .types.execution import Process
    if not isinstance(u, Handle | Socket):
        return True
    if isinstance(u, Socket):
        return False
    if u.file:
        for v in u.file.neighbors():
            if isinstance(v, Process):
                return True
    return False

def single_threads(u : Vertex) -> bool:
    from .types.execution import Thread, Process
    return not isinstance(u, Thread) or len([v for v in u.neighbors() if isinstance(v, Process)]) > 1

@cache
def single_processes(u : Vertex) -> bool:
    from .types.execution import Process
    from .types.network import Host

    if not isinstance(u, Process):
        return True

    def test_process_vertex(u : Process) -> bool:
        return bool([v for v in u.neighbors() if not isinstance(v, (Process, Host))]) or u.parent_process == None
        
    return test_process_vertex(u) or any(single_processes(v) for v in u.children_processes)
    


Key_KeyEntry_Model : System[Key | KeyEntry | Handle] = System()
_strong_keys : set[Key] = set()

@cache
def changed_registry(u : Vertex) -> bool:           # This filter is so fucking complicated!!!
    from .types.registry import Key, KeyEntry, Handle, ChangesTowards, SetsEntry, DeletesEntry, HasSubKey
    from .config import ColorMap

    if not isinstance(u, Key | KeyEntry | Handle):
        return True
    
    if isinstance(u, Handle):
        with Key_KeyEntry_Model.Entangled(u):
            return any(changed_registry(v) for v in u.neighbors() if isinstance(v, KeyEntry) if v not in Key_KeyEntry_Model) or any(changed_registry(v) for v in u.neighbors() if isinstance(v, Key) and v not in Key_KeyEntry_Model)
    
    if isinstance(u, KeyEntry):
        with Key_KeyEntry_Model.Entangled(u):              # Checking this Entry: do not do it twice!

            if u.key in _strong_keys:               # We are actually propagating the collapse of the wave function!
                return True
            
            if u.color == ColorMap.get_config().Deleted:        # This is an important Entry : it collapses the wave function!
                _strong_keys.add(u.key)
                Key_KeyEntry_Model.add_callback(_strong_keys.remove, u.key)
                return True

            for e in u.edges:
                if isinstance(e, SetsEntry | DeletesEntry):     # This is an important Entry : it collapses the wave function!
                    _strong_keys.add(u.key)
                    Key_KeyEntry_Model.add_callback(_strong_keys.remove, u.key)
                    return True
            
            for e in u.edges:                   # We got another particle that will probably collapse the wave function nearby : entangle with it!
                if isinstance(e, ChangesTowards) and e.source is u and e.destination not in Key_KeyEntry_Model:
                    return changed_registry(e.destination)
                
            if u.key not in Key_KeyEntry_Model:     # The Key has not been tested yet. Pass on to it!
                return changed_registry(u.key)

            if Key_KeyEntry_Model.population:   # No clue: pass on to someone else...
                return changed_registry(Key_KeyEntry_Model.population.pop())
            else:       # No clue and there is no one else: it means we are on a weak Key and we can start the collapse.
                return False
    
    if isinstance(u, Key):
    
        with Key_KeyEntry_Model.Entangled(u):
            Key_KeyEntry_Model.population = u.entries       # Set the populations of entries to entangle with.

            if u.color == ColorMap.get_config().Deleted:        # This is an important Key : it collapses the wave function!
                _strong_keys.add(u)
                Key_KeyEntry_Model.add_callback(_strong_keys.remove, u)

            if Key_KeyEntry_Model.population and all(changed_registry(v) for v in Key_KeyEntry_Model.include()):       # If it is a strong key, positive collapse will start in one of the entries and propagate to all others.
                return True
            
            if u in _strong_keys:       # It was a strong Key due to its own color.
                return True
            
            for h in u.neighbors():         # A handle that might have worked on important values might be working on this key because of a symbolic link.
                if isinstance(h, Handle) and h not in Key_KeyEntry_Model and changed_registry(h) and h.key is u:
                    return True
        
        if any(changed_registry(e.destination) for e in u.edges if isinstance(e, HasSubKey) and e.source is u and e.destination not in Key_KeyEntry_Model):     # If not a strong key, at least it has strong heirs?
            return True
        
        return False


significant_call_only = Filter(lambda e : True, important_call_filter)
no_data_nodes = Filter(lambda e : True, data_filter)
no_simple_imports = Filter(lambda e : True, simple_imports)
no_handle_nodes = Filter(lambda e : True, no_handles)
injected_threads_only = Filter(lambda e : True, single_threads)
significant_processes_only = Filter(lambda e : True, single_processes)
modified_registry_only = Filter(lambda e : True, changed_registry)

del important_call_filter, data_filter, simple_imports, no_handles, single_threads, Callable, Iterable, Edge, Vertex, Graph, InstanceReferencingClass, Key, KeyEntry


"""
This module defines functions that get quite useful when building MetaGraphs.
"""

import pathlib
from ..source.metagraph import MetaGraph

__all__ = ["import_env", "save", "load", "entries", "remove", "export_MG", "import_MG"]





def import_env(d : dict):

    """
    Imports all the necessary classes for building MetaGraphs
    """

    from ...bakery.source import types
    from ..source.metagraph import MetaGraph, MetaEdge, MetaArrow, MetaVertex
    from types import ModuleType

    l = locals().copy()
    if "l" in l:
        l.pop("l")
    l.pop("d")
    l.pop("types")

    for name in dir(types):
        val = getattr(types, name)
        if isinstance(val, ModuleType):
            l[name] = val

    d.update(l)





def save(mg : MetaGraph, name : str):
    """
    Saves the given MetaGraph in the library with the given name.
    """
    from ..source.metagraph import MetaGraph
    from os.path import dirname, sep
    from os import makedirs
    from pickle import dump

    if not isinstance(mg, MetaGraph) or not isinstance(name, str):
        raise TypeError("Expected MetaGraph, str, got " + repr(type(mg).__name__) + " and " + repr(type(name).__name__))
    
    path = dirname(__file__)
    if not path.endswith(sep):
        path += sep
    makedirs(path + "lib", exist_ok=True)
    path += "lib" + sep + name + ".pyt"

    with open(path, "wb") as f:
        dump(mg, f)





def load(name : str) -> MetaGraph:
    """
    Loads the MetaGraph with given name from the library.
    """
    from os.path import dirname, sep, isfile
    from os import makedirs
    from pickle import load

    if not isinstance(name, str):
        raise TypeError("Expected str, got " + repr(type(name).__name__))

    path = dirname(__file__)
    if not path.endswith(sep):
        path += sep
    makedirs(path + "lib", exist_ok=True)
    path += "lib" + sep + name + ".pyt"

    if not isfile(path):
        raise FileNotFoundError("Given MetaGraph name does not exist in the library.")
    
    with open(path, "rb") as f:
        return load(f)





def entries() -> list[str]:
    """
    Returns the names of all MetaGraphs existing in the library.
    """
    from os.path import dirname, sep, splitext, split
    from os import makedirs
    from glob import iglob

    path = dirname(__file__)
    if not path.endswith(sep):
        path += sep
    makedirs(path + "lib", exist_ok=True)
    path += "lib" + sep

    l = []
    for name in iglob(path + "**.pyt", recursive = True):
        l.append(splitext(split(name)[1])[0])

    return l





def remove(name : str):
    """
    Removes the given name from the MetaGraph library.
    """
    from os.path import dirname, sep, isfile
    from os import remove, makedirs

    if not isinstance(name, str):
        raise TypeError("Expected str, got " + repr(type(name).__name__))

    path = dirname(__file__)
    if not path.endswith(sep):
        path += sep
    makedirs(path + "lib", exist_ok=True)
    path += "lib" + sep + name + ".pyt"

    if not isfile(path):
        raise FileNotFoundError("No such MetaGraph : " + repr(path))

    remove(path)





def export_MG(mg : MetaGraph, file : str | pathlib.Path):
    """
    Exports the given MetaGraph to the given file
    """
    from pathlib import Path
    if not isinstance(mg, MetaGraph):
        raise TypeError("Expected MetaGraph, path, got " + repr(type(mg).__name__) + " and " + repr(type(file).__name__))
    if isinstance(file, str):
        try:
            file = Path(file)
        except:
            raise ValueError("Invalid path : '{}'".format(file))
    if not isinstance(file, Path):
        raise TypeError("Expected MetaGraph, path, got " + repr(type(mg).__name__) + " and " + repr(type(file).__name__))
    from pickle import dump
    with file.open("wb") as f:
        dump(mg, f)
    




def import_MG(file : str | pathlib.Path) -> MetaGraph:
    """
    Imports a MetaGraph from the given path.
    """
    from pathlib import Path
    if isinstance(file, str):
        try:
            file = Path(file)
        except:
            raise ValueError("Invalid path : '{}'".format(file))
    if not isinstance(file, Path):
        raise TypeError("Expected path, got " + repr(type(file).__name__))
    from pickle import load
    with file.open("rb") as f:
        return load(f)


import logging
from .source.build import Builder
import os

logging.getLogger().setLevel(logging.CRITICAL)

paths = [
    "./data/Cuckoo/",
]

work = set()
for p in paths:
    work.update(p + f for f in os.listdir(p))


for sample in work:
    b = Builder(sample)
    n = 0
    try:
        for c in b.calls():
            if c["api"] == "send":
                n += 1
        if n:
            print(sample, ":", n, "calls to send")
    except KeyError:
        pass


"""
This module defines all graph-related objects.
"""

from threading import RLock, Thread
from typing import Any, Callable, Dict, Iterable, Iterator, Optional, Set, Tuple, Type, TypeVar, Union
from .colors import Color
from Viper.meta.iterable import InstanceReferencingClass
from .utils import chrono

__all__ = ["Vertex", "UniqueVertex", "Edge", "Arrow", "Graph", "find_or_create"]





class Vertex(metaclass = InstanceReferencingClass):

    """
    A vertex for a graph. Can be linked to other vertices and added to a graph.
    """

    __slots__ = {
        "edges" : "The set of edges linking this vertex to others.",
        "color" : "The color to use for representation (three values between 0 and 1)",
        "parent" : "A vertex that 'contains' this one.",
        "children" : "The tuple of vertices (directly) contained by this one.",
        "graph" : "The graph object containing this vertex",
        "__size" : "The size of the vertex"
        }
    
    __pickle_slots__ = {
        "color",
        "parent",
        "size"
    }

    def __init__(self, *, c : Color | None = None, parent : Optional["Vertex"] = None) -> None:
        from .colors import Color
        from .config import ColorMap, SizeMap
        if c is None:
            c = ColorMap.get_config().get_color(type(self))
        if not isinstance(c, Color):
            raise TypeError("Expected color for c, got " + repr(c.__class__.__name__))
        if parent != None and not isinstance(parent, Vertex):
            raise TypeError("Expected vertex for parent, got " + repr(parent.__class__.__name__))
        self.edges : Set[Edge] = set()
        self.color : Color = c
        self.__size : float = SizeMap.get_config().get_size(type(self))
        self.parent : Vertex = parent
        self.graph : Graph = None
        if parent:
            parent.children += (self, )
        self.children : Tuple[Vertex] = ()
        for g in Graph.active_graphs():
            g.vertices.add(self)

    __vertexing = False
    __comparing : set[tuple["Vertex"]] = set()

    @property
    def label(self) -> str:
        """
        The label used when plotting this vertex.
        """
        return type(self).__name__

    @property
    def size(self) -> float:
        """
        The size of this Vertex. Default is 20.0.
        """
        return self.__size
    
    @size.setter
    def size(self, value : float):
        if not isinstance(value, float):
            try:
                value = float(value)
            except:
                pass
        if not isinstance(value, float):
            raise TypeError("Expected float for size, got " + repr(type(value).__name__))
        if value < 0 or value in (float("inf"), float("nan")):
            raise ValueError("Expected positive finite number for size, got " + repr(value))
        self.__size = value

    def __repr__(self) -> str:
        """
        Implements repr(self)
        """
        if Vertex.__vertexing:
            return str(type(self)) + "()"
        Vertex.__vertexing = True
        r = type(self).__name__ + "(" + ", ".join(str(name) + " = " + str(getattr(self, name)) for name in self.__slots__ if not name.startswith("_")) + ")"
        Vertex.__vertexing = False
        return r
    
    def __str__(self) -> str:
        """
        Implements str(self)
        """
        if Vertex.__vertexing:
            return str(type(self))
        Vertex.__vertexing = True
        r = "(" + ", ".join(str(name) + " = " + str(getattr(self, name)) for name in self.__slots__ if not name.startswith("_")) + ")"
        Vertex.__vertexing = False
        return r
    
    def __hash__(self) -> int:
        """
        Implements hash(self)
        """
        n = 0
        for name in dir(self):
            if not name.startswith("__") and not name.endswith("__") and name not in Vertex.__slots__:
                try:
                    n += hash(name) * hash(getattr(self, name))
                except:
                    pass
        return n
    
    def __eq__(self, o: object) -> bool:
        """
        Implements self == o
        """
        if self is o:
            return True
        if (self, o) in self.__comparing or (o, self) in self.__comparing:
            return True
        self.__comparing.add((self, o))
        if type(self) != type(o):
            return False
        res = all(getattr(self, name) == getattr(o, name) for name in type(self).__slots__)
        self.__comparing.remove((self, o))
        return res
    
    def __reduce__(self) -> str | tuple[Any, ...]:
        """
        Implements dumping of self
        """
        # Note : When serializing a vertex, the edges/arrows should not be serialized. These will be handled independently when serializing a graph.
        func, args, state = super().__reduce__()
        state : dict
        if "edges" in state or "graph" in state:
            raise RuntimeError("Cannot pickle edges or graph of a vertex")
        return type(self), (), state
    
    def __setstate__(self, state : Dict[str, Any]):
        """
        Implements loading of self
        """
        for k, v in state.items():
            setattr(self, k, v)
    
    def __getstate__(self) -> Dict[str, Any]:
        """
        Implements dumping of self
        """
        from .utils import extract_pickle_slots
        return {name : getattr(self, name) for name in extract_pickle_slots(type(self))}
    
    def __copy__(self) -> "Vertex":
        """
        Implements copy of self
        """
        from inspect import ismethod
        cp = type(self)()
        cp.color = self.color
        for name in dir(self):
            if not name.startswith("_") and not name.endswith("_") and name not in Vertex.__slots__ and not ismethod(getattr(self, name)):
                setattr(cp, name, getattr(self, name))
        return cp
    
    def __deepcopy__(self, memo : Dict[int, Any]) -> "Vertex":
        """
        Implements deepcopy of self
        """
        from copy import deepcopy
        from inspect import ismethod
        cp = type(self)()
        cp.color = self.color
        memo[id(self)] = cp
        for name in dir(self):
            if not name.startswith("_") and not name.endswith("_") and name not in Vertex.__slots__ and not ismethod(getattr(self, name)):
                setattr(cp, name, deepcopy(getattr(self, name), memo))
        return cp
                    
    def neighbors(self) -> Iterator["Vertex"]:
        """
        Iterates over all the neighbor vertices.
        """
        for e in self.edges:
            if e.source is self:
                yield e.destination
            else:
                yield e.source
    
    def outwards(self) -> Iterator["Arrow"]:
        """
        Iterates over the outwards neighbors of this vertex (neighbors linked by an outgoing arrow).
        """
        for e in self.edges:
            if isinstance(e, Arrow) and e.source is self:
                yield e.destination
    
    def inwards(self) -> Iterator["Arrow"]:
        """
        Iterates over the inwards neighbors of this vertex (neighbors linked by an incomming arrow).
        """
        for e in self.edges:
            if isinstance(e, Arrow) and e.destination is self:
                yield e.source
    
    def linked(self) -> Iterator["Edge"]:
        """
        Iterates over the undirected neighbors if this vertex (neighbors linked by a strict edge).
        """
        for e in self.edges:
            if not isinstance(e, Arrow):
                yield (e.source if e is not e.source else e.destination)

    def connect(self, o : "Vertex", name : str, *, directional : bool = False, c : Optional[Color] = None) -> None:
        """
        Links this vertex to another. Directional indicates if the link should be an arrow instead of an edge.
        """
        if not isinstance(o, Vertex) or not isinstance(name, str):
            raise TypeError("Expected vertex, str, got " + repr(o.__class__.__name__) + " and " + repr(name.__class__.__name__))
        if not isinstance(directional, bool):
            raise TypeError("Expected bool for directional, got" + repr(directional.__class__.__name__))
        if c == None:
            c = Color.average(self.color, o.color)
        if not isinstance(c, Color):
            raise TypeError("Expected color for c, got " + repr(c.__class__.__name__))
        if directional:
            e = Arrow(name, self, o, c=c)
        else:
            e = Edge(name, self, o, c=c)
        e.write()

    @classmethod
    @chrono
    def add_vertices_to_graph(cls : Type["Vertex"], G : "Graph", fil : Optional[Callable[["Vertex"], bool]] = None):
        """
        Adds all vertices of this class to a graph.
        If given a filter function fil, only filtered vertices will be added.
        """
        if not isinstance(G, Graph):
            raise TypeError("Expected graph, got " + repr(G.__class__.__name__))
        if fil != None and not callable(fil):
            raise TypeError("Expected callable for filter, got " + repr(fil.__class__.__name__))
        if fil == None:
            G.vertices.update(cls)
        else:
            G.vertices.update(filter(fil, cls))
                




class UniqueVertex(Vertex):

    """
    Vertices from this class all have different hashes, (but can still compare equals if they have equal custom attributes).
    """

    __N = 0          # Current index
    __lock = RLock() # The lock used to ensure instance creation is thread-safe

    __slots__ = {
        "__id" : "The unique identifier of this vertex."
    }

    def __init__(self, *, c: Color | None = None, parent: Optional["Vertex"] = None) -> None:
        self.__id = None
        self.__pickID()
        super().__init__(c = c, parent = parent)

    def __pickID(self):
        """
        Picks a unique id for self if it doesn't have one.
        """
        if self.__id == None:
            with UniqueVertex.__lock:
                self.__id = UniqueVertex.__N
                UniqueVertex.__N += 1
    
    def __hash__(self) -> int:
        """
        Implements hash(self).
        """
        self.__pickID()
        return self.__id
    
    def __eq__(self, o: object) -> bool:
        """
        Implements self == o.
        """
        return type(self) == type(o) and self.__id == o.__id
            




class Edge(metaclass = InstanceReferencingClass):

    """
    An (undirected) edge for a graph. Links two vertices together.
    """

    __slots__ = {
        "source" : "The source vertex.",
        "destination" : "The destination vertex.",
        "color" : "The color to use for representation (all between 0 and 1)",
        "__weight" : "The weight of the edge"
    }

    __pickle_slots__ = {
        "color",
        "source",
        "destination",
        "weight"
    }

    def __init__(self, source : Vertex, destination : Vertex, *, c: Color | None = None, auto_write : bool = True) -> None:
        from .colors import Color
        if not isinstance(source, Vertex) or not isinstance(destination, Vertex):
            raise TypeError("Expected vertex, vertex, got " + repr(source.__class__.__name__) + " and " + repr(destination.__class__.__name__))
        if c == None:
            from .config import ColorMap
            if ColorMap.get_config().has_color(type(self)):
                c = ColorMap.get_config().get_color(type(self))
            else:
                c = Color.average(source.color, destination.color)
        if not isinstance(c, Color):
            raise TypeError("Expected color for c, got " + repr(c.__class__.__name__))
        if not isinstance(auto_write, bool):
            raise TypeError("Expected bool for write, got " + repr(auto_write.__class__.__name__))
        self.source : Vertex = source
        self.destination : Vertex = destination
        self.color : Color = c
        self.__weight : float = 1.
        if auto_write:
            self.write()
        for g in Graph.active_graphs():
            g.edges.add(self)
    
    @property
    def label(self) -> str:
        """
        The label used when plotting this edge.
        """
        return type(self).__name__
    
    @property
    def weight(self) -> float:
        """
        The weight of this Edge. Default is 1.0.
        """
        return self.__weight
    
    @weight.setter
    def weight(self, value : float):
        if not isinstance(value, float):
            try:
                value = float(value)
            except:
                pass
        if not isinstance(value, float):
            raise TypeError("Expected float for weight, got " + repr(type(value).__name__))
        if value < 0 or value in (float("inf"), float("nan")):
            raise ValueError("Expected positive finite number for weight, got " + repr(value))
        self.__weight = value
    
    def __repr__(self) -> str:
        """
        Implements repr(self)
        """
        return "edge(" + repr(self.source) + ", " + repr(self.destination) + ")"
    
    def __str__(self) -> str:
        """
        Implements str(self)
        """
        return str(self.source) + " -- " + str(self.destination)
    
    def __hash__(self) -> int:
        """
        Implements hash(self)
        """
        return hash(hash(self.source) + hash(type(self)) + hash(self.destination))
    
    def __eq__(self, o: object) -> bool:
        """
        Implements self == o
        """
        if not isinstance(o, type(self)):
            return False
        return (self.source == o.source and self.destination == o.destination) or (self.source == o.destination and self.destination == o.source)
        
    def __setstate__(self, state : Dict[str, Any]):
        """
        Implements loading of self
        """
        for k, v in state.items():
            setattr(self, k, v)
    
    def __getstate__(self) -> Dict[str, Any]:
        """
        Implements dumping of self
        """
        from .utils import extract_pickle_slots
        return {name : getattr(self, name) for name in extract_pickle_slots(type(self))}
    
    def __copy__(self) -> "Edge":
        """
        Implements copy(self)
        """
        # TO-DO
        
    def write(self):
        """
        Writes this edge in the edges sets of both vertices.
        """
        self.source.edges.add(self)
        self.destination.edges.add(self)
        if self.source.graph:
            self.source.graph.edges.add(self)
        elif self.destination.graph:
            self.destination.graph.edges.add(self)
    
    def delete(self) -> Tuple[Vertex, Vertex]:
        """
        Deletes the link. (Deletes it from the vertices egde sets)
        """
        self.source.edges.discard(self)
        self.destination.edges.discard(self)
        if self.source.graph:
            self.source.graph.edges.discard(self)
        elif self.destination.graph:
            self.destination.graph.edges.discard(self)
    
    @classmethod
    @chrono
    def add_edges_to_graph(cls : Type["Edge"], G : "Graph", fil : Optional[Callable[["Edge"], bool]] = None):
        """
        Adds all edges of this class to a graph.
        If given a filter function fil, only filtered edges will be added.
        """
        if not isinstance(G, Graph):
            raise TypeError("Expected graph, got " + repr(G.__class__.__name__))
        if fil != None and not callable(fil):
            raise TypeError("Expected callable for filter, got " + repr(fil.__class__.__name__))
        if fil == None:
            G.edges.update(cls)
        else:
            G.edges.update(filter(fil, cls))

    
        


class Arrow(Edge):

    """
    An arrow for a graph. Links two vertices together.
    """
    
    def __repr__(self) -> str:
        """
        Implements repr(self).
        """
        return "arrow(" + repr(self.source) + ", " + repr(self.destination) + ")"
    
    def __str__(self) -> str:
        """
        Implements str(self).
        """
        return str(self.source) + " --> " + str(self.destination)
    
    def __hash__(self) -> int:
        """
        Implements hash(self).
        """
        return hash(-hash(self.source) + hash(type(self)) + hash(self.destination))
    
    def __eq__(self, o: object) -> bool:
        """
        Implements self == o.
        """
        if not isinstance(o, type(self)):
            return False
        return self.source == o.source and self.destination == o.destination
    





CLS = TypeVar("CLS", bound=Vertex)

@chrono
def find_or_create(cls : Type[CLS], **kwargs) -> Tuple[CLS, bool]:
    """
    Finds in the given class cls for a vertex which attributes have values matching the given keyword arguments.
    If there is one, returns it. Otherwise, creates it.
    Returns the object and a boolean indicating if the object existed before.
    """
    for ist in cls:
        if all(hasattr(ist, name) and getattr(ist, name) == value for name, value in kwargs.items()):
            return ist, True
    ist = cls()
    for name, value in kwargs.items():
        setattr(ist, name, value)
    return ist, False





class Graph:

    """
    A graph class. Add starting vertices and use explore() to complete the graph.
    Graph can be used in context managers to append all Vertices and Edges created in the current thread to the context Graph.
    """

    __slots__ = {
        "vertices" : "The set of vertices in this graph",
        "edges" : "The set of edges in this graph",
        "data" : "A dictionary holding additional data for this graph"
    }

    __hash__ = None

    __active_graphs : dict[Thread, list["Graph"]] = {}

    def __init__(self) -> None:
        from typing import Any
        self.vertices : Set[Vertex] = set()
        self.edges : Set[Edge] = set()
        self.data : dict[str, Any] = {}
    
    def __enter__(self):
        """
        Implements with self.
        """
        from threading import current_thread
        if current_thread() not in Graph.__active_graphs:
            Graph.__active_graphs[current_thread()] = []
        Graph.__active_graphs[current_thread()].append(self)
    
    def __exit__(self, exc_type, exc_value, traceback):
        """
        Implements with self.
        """
        from threading import current_thread
        Graph.__active_graphs[current_thread()].remove(self)
    
    @staticmethod
    def active_graphs() -> list["Graph"]:
        """
        Returns the list of all the active Graphs in the current thread.
        """
        from threading import current_thread
        if current_thread() not in Graph.__active_graphs:
            return []
        return Graph.__active_graphs[current_thread()].copy()

    def add_vertex(self, v : Vertex, explore : bool = True) -> None:
        """
        Adds a vertex to the graph and explores from this vertex.
        """
        if not isinstance(v, Vertex):
            raise TypeError("Expected vertex, got " + repr(v.__class__.__name__))
        if not isinstance(explore, bool):
            raise TypeError("Expected bool for explore, got " + repr(explore.__class__.__name__))
        if v not in self.vertices:
            self.vertices.add(v)
            v.graph = self
            if explore:
                self.explore(v)
    
    @chrono
    def explore(self, source : Optional[Vertex] = None) -> None:
        """
        Explores the graph for more linked vertices. If a vertex of the graph is given, only searches starting from that vertex. Otherwise, searches from all vertices.
        """
        if source == None:
            to_explore = self.vertices.copy()
        else:
            if not isinstance(source, Vertex):
                raise TypeError("Expected vertex, got " + repr(source.__class__.__name__))
            if source not in self.vertices:
                raise KeyError("Vertex not in graph.")
            to_explore = {source}
        from itertools import chain
        seen_vertices = set()
        seen_edges = set()
        while to_explore:
            u = to_explore.pop()
            seen_vertices.add(u)
            seen_edges.update(u.edges)
            u.graph = self
            next_vertices = chain(u.neighbors(), iter(u.children))
            if u.parent:
                next_vertices = chain(next_vertices, (u.parent, ))
            new = set()
            for v in next_vertices:
                if v not in seen_vertices:
                    new.add(v)
            to_explore.update(new)
        self.vertices.update(seen_vertices)
        self.edges.update(seen_edges)

    def pairs(self) -> Iterator[tuple[Vertex, Edge, Vertex]]:
        """
        Yields all (vertex u, edge e, vertex v) tuples where (u, v) is a pair of linked vertices. e may be an edge or an arrow directed from u to v.
        """
        for e in self.edges:
            yield e.source, e, e.destination

    def __len__(self) -> int:
        """
        Implements len(self). Returns the number of vertices.
        """
        return len(self.vertices)
    
    @property
    def n(self) -> int:
        """
        The number of vertices.
        """
        return len(self)
    
    @property
    def m(self) -> int:
        """
        The number of edges and arrows.
        """
        return len(self.edges)
    
    def diameter(self) -> int:
        """
        Returns the diameter of the graph.
        """
        import networkx as nx
        G = self.NetworkX().to_undirected()
        return nx.diameter(G)
    
    def append(self, value : Vertex | Edge, explore : bool = False):
        """
        Adds a vertex or an edge to the graph.
        If explore is True, explores the graph from the added vertex or the source vertex of the added edge.
        """
        if not isinstance(explore, bool):
            raise TypeError("Expected bool for explore, got " + repr(explore.__class__.__name__))
        if isinstance(value, Vertex):
            self.vertices.add(value)
            value.graph = self
            if explore:
                self.explore(value)
        elif isinstance(value, Edge):
            value.write()
            self.edges.add(value)
            if explore:
                self.explore(value.source)
        else:
            raise TypeError("Expected edge or vertex, got " + repr(value.__class__.__name__))
    
    def remove(self, value : Vertex | Edge):
        """
        Removes a vertex or an edge from the graph.
        When removing a vertex, it will also remove all edges/arrows connected to it.
        """
        if isinstance(value, Vertex):
            self.edges.difference_update(value.edges)
            for e in value.edges.copy(): 
                e.source.edges.discard(e)
                e.destination.edges.discard(e)
            self.vertices.discard(value)
        elif isinstance(value, Edge):
            self.edges.discard(value)
            value.source.edges.discard(value)
            value.destination.edges.discard(value)
        else:
            raise TypeError("Expected edge or vertex, got " + repr(type(value).__name__))

    @chrono
    def extend(self, values : Iterable[Vertex | Edge], explore : bool = False):
        """
        Extends the graph with an iterable of vertices and/or edges.
        """
        from typing import Iterable
        if not isinstance(explore, bool):
            raise TypeError("Expected bool for explore, got " + repr(explore.__class__.__name__))
        if not isinstance(values, Iterable):
            raise TypeError("Expected iterable, got " + repr(values.__class__.__name__))
        for value in values:
            if isinstance(value, Vertex):
                self.vertices.add(value)
                value.graph = self
                if explore:
                    self.explore(value)
            elif isinstance(value, Edge):
                value.write()
                self.edges.add(value)
                if explore:
                    self.explore(value.source)
            else:
                raise TypeError("Expected edge or vertex, got " + repr(value.__class__.__name__))
        
    def paint(self, c : Color):
        """
        Changes the colors of all vertices and edges/arrows of this graph to the given Color.
        """
        from .colors import Color
        if not isinstance(c, Color):
            raise TypeError("Expected Color, got " + repr(type(c).__name__))
        for u in self.vertices:
            u.color = c
        for e in self.edges:
            e.color = c

    def __getstate__(self) -> Dict[str, Any]:
        """
        Implements dumping of self.
        """
        return {
            "vertices" : self.vertices,
            "edges" : self.edges, 
            "data" : self.data
        }

    def __setstate__(self, state : Dict[str, Any]):
        """
        Implements loading of self.
        """
        for name, value in state.items():
            setattr(self, name, value)
        for v in self.vertices:
            v.graph = self
        for e in self.edges:
            e.write()
   
    # def __copy__(self) -> "graph":
    #     """
    #     Implements copy of self
    #     """
    #     from copy import copy
    #     translation = {u : copy(u) for u in self.vertices}
    #     cp = graph()
    #     cp.vertices.update(translation.values())
    #     for u, e, v in self.pairs():
    #         e = copy(e)
    #         e.source = translation[u]
    #         e.destination = translation[v]
    #         e.write()
    #     return cp
    
    # def __deepcopy__(self, memo : Dict[int, Any]) -> "graph":
    #     """
    #     Implements deepcopy of self
    #     """
    #     from copy import deepcopy
    #     translation = {u : deepcopy(u, memo) for u in self.vertices}
    #     cp = graph()
    #     cp.vertices.update(translation.values())
    #     for u, e, v in self.pairs():
    #         e = deepcopy(e, memo)
    #         e.source = translation[u]
    #         e.destination = translation[v]
    #         e.write()
    #     return cp

    @chrono
    def export(self, file : str, *, subgraph_supported : bool = False) -> None:
        """
        Writes this graph under the GEXF format into file.
        """
        # List of possible attributes to include in visuals :
        # Node Size : proportional to radius! not surface area!
        # Color : R, G, B
        # Node Shape : Any of "disc", "square", "triangle" or "diamond"
        # Edge Thickness : same as weight?
        # Edge Shape : Any of "solid", "dotted", "dashed" or "double"
        from datetime import date
        import json

        forbidden_characters : dict[int, str] = {i : "�" for i in range(32) if chr(i) not in {"\n", "\t", "\r"}}

        if not isinstance(subgraph_supported, bool):
            raise TypeError("Expected bool for subgraph_supported, got " + repr(subgraph_supported.__class__.__name__))
        try:
            with open(file, "wb") as f:
                import xml.etree.ElementTree as ET
                head = b'<?xml version="1.0" encoding="UTF-8"?>'
                root = ET.Element("gexf", attrib={"xmlns":"http://www.gexf.net/1.3", "version":"1.3", "xmlns:viz":"http://www.gexf.net/1.3/viz", "xmlns:xsi":"http://www.w3.org/2001/XMLSchema-instance", "xsi:schemaLocation":"http://www.gexf.net/1.3 http://www.gexf.net/1.3/gexf.xsd"})
                meta = ET.SubElement(root, "meta", lastmodifieddate=date.today().isoformat())
                creator = ET.SubElement(meta, "creator")
                creator.text = "Graph Builder"
                description = ET.SubElement(meta, "description")
                description.text = ""
                graph = ET.SubElement(root, "graph", mode="static")
                node_attr = ET.SubElement(graph, "attributes", **{"class" : "node", "mode" : "static"})
                node_attributes = {}
                edge_attr = ET.SubElement(graph, "attributes", **{"class" : "edge", "mode" : "static"})
                edge_attributes = {}
                nodes = ET.SubElement(graph, "nodes")
                edges = ET.SubElement(graph, "edges")
                n_ids = {u : i for i, u in enumerate(self.vertices)}
                additional_links : Set[Tuple[Vertex, Vertex]] = set()
                for u, i in n_ids.items():
                    d = {name : str(getattr(u, name)).translate(forbidden_characters) for name in u.__pickle_slots__ if not name.startswith("_")}
                    d["Type"] = type(u).__name__
                    if "__weakref__" in d:
                        d = {}
                    if u.parent:
                        if subgraph_supported:
                            node_i = ET.SubElement(nodes, "node", id=str(i), label=u.label.translate(forbidden_characters), pid=str(n_ids[u.parent]))
                        else:
                            node_i = ET.SubElement(nodes, "node", id=str(i), label=u.label.translate(forbidden_characters))
                            additional_links.add((u.parent, u))
                    else:
                        node_i = ET.SubElement(nodes, "node", id=str(i), label=u.label.translate(forbidden_characters))
                    attr_i = ET.SubElement(node_i, "attvalues")
                    for name, value in d.items():
                        if name not in node_attributes:
                            node_attributes[name] = ET.SubElement(node_attr, "attribute", id=name, title=name, type="string")
                        if isinstance(value, (dict, list)):
                            try:
                                svalue = json.dumps(value, indent = "\t")
                            except:
                                svalue = str(value)
                        else:
                            svalue = str(value)
                        ET.SubElement(attr_i, "attvalue", **{"for" : name, "value" : svalue.translate(forbidden_characters)})
                    size = ET.SubElement(node_i, "viz:size", value=str(u.size))
                    color = ET.SubElement(node_i, "viz:color", r=str(int(u.color.R * 255)), g=str(int(u.color.G * 255)), b=str(int(u.color.B * 255)))
                for u, e, v in self.pairs():
                    d = {name : str(getattr(e, name)).translate(forbidden_characters) for name in e.__pickle_slots__ if not name.startswith("_")}
                    d["Type"] = type(e).__name__
                    if "__weakref__" in d:
                        d = {}
                    try:
                        edge_i = ET.SubElement(edges, "edge", source=str(n_ids[u]), target=str(n_ids[v]), type=("directed" if isinstance(e, Arrow) else "undirected"), label=e.label.translate(forbidden_characters), attrib=d, weight=str(e.weight))
                    except KeyError:
                        raise
                    color = ET.SubElement(edge_i, "viz:color", r=str(int(e.color.R * 255)), g=str(int(e.color.G * 255)), b=str(int(e.color.B * 255)))
                    attr_i = ET.SubElement(edge_i, "attvalues")
                    for name, value in d.items():
                        if name not in edge_attributes:
                            edge_attributes[name] = ET.SubElement(edge_attr, "attribute", id=name, title=name, type="string")
                        if isinstance(value, (dict, list)):
                            try:
                                svalue = json.dumps(value, indent = "\t")
                            except:
                                svalue = str(value)
                        else:
                            svalue = str(value)
                        ET.SubElement(attr_i, "attvalue", **{"for" : name, "value" : svalue.translate(forbidden_characters)})
                for u, v in additional_links:
                    edge_i = ET.SubElement(edges, "edge", source=str(n_ids[u]), target=str(n_ids[v]), type="directed", label="contains", weight="1.0")
                f.write(head + b"\n")
                ET.indent(root, "\t")
                for line in ET.tostringlist(root):
                    f.write(line + b"\n")
        except Exception as E:
            raise

    @chrono
    def NetworkX(self) -> Union["nx.Graph", "nx.DiGraph", "nx.MultiGraph", "nx.MultiDiGraph"]:
        """
        Returns a NeworkX version of this graph.
        """
        import networkx as nx
        multi = False
        directed = False
        seen : Set[Tuple[Vertex, Vertex]] = set()
        for u, e, v in self.pairs():
            if (u, v) in seen:
                multi = True
            if isinstance(e, Arrow):
                directed = True
            if multi and directed:
                break
        if multi and directed:
            G = nx.MultiDiGraph()
        elif multi:
            G = nx.MultiGraph()
        elif directed:
            G = nx.DiGraph()
        else:
            G = nx.Graph()
        for u, e, v in self.pairs():
            G.add_edge(u, v)        # à améliorer
        for u in self.vertices:
            for v in u.children:
                G.add_edge(u, v)
        return G
    

del RLock, Any, Callable, Dict, Iterable, Iterator, Optional, Set, Tuple, Type, TypeVar, Union, Color, InstanceReferencingClass, chrono, CLS


"""
This module adds a strong logging system to the bakery package.
"""

import logging
import sys

__all__ = ["logger", "stdout_handler", "standart_formatter", "set_level", "create_log_file"]





logger = logging.getLogger()
logger.setLevel(0)
stdout_handler = logging.StreamHandler(sys.stdout)
standart_formatter = logging.Formatter("%(levelname)-8s : %(module)-10s : %(asctime)-24s : %(message)s")

logger.addHandler(stdout_handler)
stdout_handler.setFormatter(standart_formatter)





def set_level(level : int):
    """
    Sets the logging level to stdout.
    """
    stdout_handler.setLevel(level)

def create_log_file(path : str, level : int = logging.DEBUG):
    """
    Creates a new log file to which logs with a level higher or equal to the current level will be written.
    """
    handler = logging.FileHandler(path, "w")
    handler.setFormatter(standart_formatter)
    handler.setLevel(level)
    logger.addHandler(handler)

set_level(logging.WARNING)

del logging, sys


"""
This module defines the Vertices and Edges that are related to filesystem activity.
"""

from ...logger import logger
from typing import Optional, Type
from ..colors import Color
from ..graph import Arrow, Edge, UniqueVertex, Vertex, Graph
from .execution import Call, CallHandler, Process
from .network import Host
from ..utils import chrono
from ..event import Event
from ..build import BuildingPhase
import pathlib

__all__ = ["File", "NewFile", "Directory", "Handle", "UsesFile", "UsesDirectory", "Opens", "Closes", "CreatesFile", "CreatesDirectory", "DeletesFile", "DeletesDirectory", "Reads", "Writes", "Conveys", "HasDrive", "HasHandle", "Contains", "IsCopyiedInto"]





logger.debug("Loading file system library.")

class File(UniqueVertex):

    """
    A file vertex. Represents a (real) file opened during execution.
    """

    __slots__ = {
        "__path" : "The path to the file in the file system",
    }

    __pickle_slots__ = {
        "path"
    }

    def __init__(self, *, c: Color | None = None, parent: Optional[Vertex] = None) -> None:
        from pathlib import PurePath
        super().__init__(c=c, parent=parent)
        self.__path : PurePath | None = None

    @property
    def path(self) -> pathlib.PurePath:
        """
        The absolute path to the file.
        """
        if self.__path is None:
            raise RuntimeError("Got a File without path.")
        return self.__path

    @path.setter
    def path(self, value : str):
        from ..utils import path_factory
        self.__path = path_factory(value)
    
    @property
    def name(self) -> str:
        """
        The name of the file (tail of the path).
        """
        if self.__path is None:
            raise RuntimeError("Got a File without path.")
        return self.__path.name
    
    @property
    def extension(self) -> str:
        """
        The file extention (lowercased without '.').
        """
        if self.__path is None:
            raise RuntimeError("Got a File without path.")
        return self.__path.suffix.lower().replace(".", "")
    
    @property
    def label(self) -> str:
        """
        Returns a label for this File node.
        """
        return 'File "' + self.name + '"'
    




class NewFile(Event):

    """
    This is an Event thrown when a File Vertex is created.
    """

    __slots__ = {
        "file" : "The File Vertex of the file that was created"
    }

    def __init__(self, file : File) -> None:
        self.file = file





class Directory(UniqueVertex):

    """
    A directory vertex. Represents a (real) directory opened during execution.
    """

    __slots__ = {
        "__path" : "The path to the directory in the file system",
    }

    __pickle_slots__ = {
        "path"
        }

    def __init__(self, *, c: Color | None = None, parent: Optional[Vertex] = None) -> None:
        from pathlib import PurePath
        super().__init__(c=c, parent=parent)
        self.__path : PurePath | None = None

    @property
    def path(self) -> pathlib.PurePath:
        """
        The absolute path to the directory.
        """
        if self.__path is None:
            raise RuntimeError("Got a File without path.")
        return self.__path

    @path.setter
    def path(self, value : str):
        from ..utils import path_factory
        self.__path = path_factory(value)
    
    @property
    def name(self) -> str:
        """
        The name of the directory (tail of the path).
        """
        if self.__path is None:
            raise RuntimeError("Got a File without path.")
        return self.__path.name if self.__path.name else self.__path.drive
    
    @property
    def label(self) -> str:
        """
        Returns a label for this Directory node.
        """
        if self.__path is None:
            raise RuntimeError("Got a File without path.")
        if not self.__path.name:
            return 'Drive "' + self.name.replace(":", "") + '"'
        return 'Directory "' + self.name + '"'
    




class Handle(UniqueVertex):

    """
    A handle vertex. Represents a file handle, used when a program opens a file.
    """

    __slots__ = {
        # For all objects:
        "synchronize" : "If the handle can perform synchronization operations on the object",

        # For file objects:
        "read" : "If the handle has the right to read the file",
        "write" : "If the handle has the right to write to the file",
        "append" : "If the handle has the right to append to the file",
        "execute" : "If the handle has the right to execute the file",
        "read_attributes" : "If the handle has the right to read file attributes",
        "write_attributes" : "If the handle can change the file attributes",
        "read_extended_attributes" : "If the handle can read the extended file attributes",
        "write_extended_attributes" : "If the handle can change the extended file attributes",

        # For directory objects:
        "list_directory" : "If the handle has the right to list the items in the given directory",
        "traverse" : "If the handle has the right to traverse the directory and use it to access subforlders/files"
    }

    __pickle_slots__ = {
        "synchronize",
        "read",
        "write",
        "append",
        "execute",
        "read_attributes",
        "write_attributes",
        "read_extended_attributes",
        "write_extended_attributes",
        "list_directory",
        "traverse"
    }

    def __init__(self, *, c: Color | None = None, parent: Optional[Vertex] = None) -> None:
        super().__init__(c=c, parent=parent)

        self.synchronize : bool = False

        self.read : bool = False
        self.write : bool = False
        self.append : bool = False
        self.execute : bool = False
        self.read_attributes : bool = False
        self.write_attributes : bool = False
        self.read_extended_attributes : bool = False
        self.write_extended_attributes : bool = False

        self.list_directory : bool = False
        self.traverse : bool = False
    
    @property
    def file(self) -> File | Directory:
        """
        Returns the file (or directory) node that this handle is working on.
        """
        for e in self.edges:
            if isinstance(e, UsesFile | UsesDirectory):
                return e.destination
        raise RuntimeError("Got a Handle working on no File or Directory.")
    




class UsesFile(Edge):

    """
    This kind of edge indicates that a handle uses a file.
    """

    label : str = ""

    source : Handle
    destination : File





class UsesDirectory(Edge):

    """
    This kind of edge indicates that a handle uses a directory.
    """

    label : str = ""

    source : Handle
    destination : Directory





class Opens(Edge):

    """
    This kind of edge indicates that a system call opened a handle.
    """

    label : str = ""

    source : Call
    destination : Handle





class Closes(Edge):

    """
    This kind of edge indicates that a system call closed a handle.
    """

    label : str = ""

    source : Call
    destination : Handle





class CreatesFile(Edge):

    """
    This kind of edge indicates that a system call created a file through a handle.
    """

    label : str = ""

    source : Call
    destination : Handle





class CreatesDirectory(Edge):

    """
    This kind of edge indicates that a system call created a directory through a handle.
    """

    label : str = ""

    source : Call
    destination : Handle





class DeletesFile(Edge):

    """
    This kind of edge indicates that a system call deleted a file through a handle.
    """

    label : str = ""

    source : Call
    destination : Handle





class DeletesDirectory(Edge):

    """
    This kind of edge indicates that a system call deleted a directory through a handle.
    """

    label : str = ""

    source : Call
    destination : Handle





class Reads(Arrow):

    """
    This kind of arrow indicates that a system call read data from a file.
    """

    from typing import TYPE_CHECKING
    if TYPE_CHECKING:
        from .data import Data
    del TYPE_CHECKING

    source : "Data"
    destination : Call





class Writes(Arrow):

    """
    This kind of arror indicates that a system call wrote data to a file.
    """

    from typing import TYPE_CHECKING
    if TYPE_CHECKING:
        from .data import Data
    del TYPE_CHECKING

    source : Call
    destination : "Data"





class Conveys(Arrow):

    """
    This kind of arrow indicates a data flux with a file handle.
    """

    label : str = ""

    from typing import TYPE_CHECKING
    if TYPE_CHECKING:
        from .data import Data
    del TYPE_CHECKING

    source : "Handle | Data"
    destination : "Handle | Data"





class HasDrive(Edge):

    """
    This kind of edge indicates that a machine owns a disk drive/partition(/root directory).
    """

    label : str = ""

    source : Host
    destination : Directory





class HasHandle(Edge):

    """
    This kind of edge indicates that a process owns a file handle.
    """

    label : str = ""

    source : Process
    destination : Handle





class Contains(Arrow):

    """
    This kind of arrow indicates that a directory contains a file or another directory.
    """

    label : str = ""

    source : Directory
    destination : Directory | File





class IsCopyiedInto(Arrow):

    """
    This kind indicates that the destination file is a copy of the source file.
    """

    source : File
    destination : File





namespace = locals().copy()
_file_to_dir_cls_table : dict[Type[Arrow | Edge], Type[Arrow | Edge]] = {}
for name, cls in filter(lambda c : isinstance(c[1], type) and issubclass(c[1], Arrow), namespace.items()):
    if name.endswith("File") and name[:-4] + "Directory" in namespace:
        _file_to_dir_cls_table[cls] = namespace[name[:-4] + "Directory"]

_active_handles : dict[Handle, str] = {}
_inverted_handles : dict[str, Handle] = {}
_existing_files : dict[str, File] = {}
_all_files : dict[str, list[File]] = {}

def declare_existing_file(path : str):
    """
    Declares an existing file in the execution environment.
    """
    from ..utils import is_path
    if not isinstance(path, str):
        raise TypeError("Expected str, got " + repr(type(path).__name__))
    if not is_path(path):
        raise ValueError("Expected a valid file path for the execution platform, got " + repr(path))
    f = File()
    f.path = path
    if path not in _all_files:
        _all_files[path] = []
    _all_files[path].append(f)
    _existing_files[path] = f
    NewFile(f).throw()
    

@chrono
def integrate_file_creation(c : Call):
    """
    Creates a File/Directory vertex if it does not exists.
    """
    from ...logger import logger
    if c.status == 1:
        logger.debug("File created.")
        if c.arguments.file_handle in _inverted_handles:
            h = _inverted_handles[c.arguments.file_handle]
        else:
            h = Handle()
            HasHandle(c.thread.process, h)
        if c.arguments.filepath in _existing_files:
            f = _existing_files[c.arguments.filepath]
            existed = True
        else:
            f = File()
            f.path = c.arguments.filepath
            existed = False
        CreatesFile(c, h)
        if not f in set(h.neighbors()):
            UsesFile(h, f)
        _inverted_handles[c.arguments.file_handle] = h
        _active_handles[h] = c.arguments.file_handle
        _existing_files[c.arguments.filepath] = f
        if c.arguments.filepath not in _all_files:
            _all_files[c.arguments.filepath] = []
        if not existed:
            _all_files[c.arguments.filepath].append(f)
        
        NewFile(f).throw()
        # if "FILE_LIST_DIRECTORY"
        # if "DELETE"
        # if "WRITE_DAC"
        # if "FILE_WRITE_ATTRIBUTES"
        # if "READ_CONTROL"
        # if "FILE_READ_EA"
        # if "FILE_WRITE_DATA"
        # if "FILE_WRITE_EA"
        # if "FILE_APPEND_DATA"
        # if "FILE_READ_ATTRIBUTES"
        # if "GENERIC_WRITE"
        # if "SYNCHRONIZE"
        # if "GENERIC_ALL"
        # if "FILE_READ_DATA" in c.flags.desired_access.split("|"):
        #     h.read = True
        # if "FILE_READ"

@chrono
def integrate_file_opening(c : Call):
    """
    Creates an File/Directory vertex.
    """
    from ...logger import logger
    if c.status == 1:
        logger.debug("File opened.")
        if c.arguments.file_handle in _inverted_handles:
            h = _inverted_handles[c.arguments.file_handle]
        else:
            h = Handle()
            HasHandle(c.thread.process, h)
        if c.arguments.filepath in _existing_files:
            f = _existing_files[c.arguments.filepath]
            existed = True
        else:
            f = File()
            f.path = c.arguments.filepath
            existed = False
        Opens(c, h)
        if not f in set(h.neighbors()):
            UsesFile(h, f)
        _inverted_handles[c.arguments.file_handle] = h
        _active_handles[h] = c.arguments.file_handle
        _existing_files[c.arguments.filepath] = f
        if c.arguments.filepath not in _all_files:
            _all_files[c.arguments.filepath] = []
        if not existed:
            _all_files[c.arguments.filepath].append(f)

        NewFile(f).throw()

@chrono
def integrate_file_closing(c : Call):
    """
    Connects a closing system call to the corresponding Handle.
    """
    from ...logger import logger
    if c.status == 1 and c.arguments.handle in _inverted_handles:
        logger.debug("File closed.")
        h = _inverted_handles.pop(c.arguments.handle)
        _active_handles.pop(h)
        Closes(c, h)

@chrono
def integrate_file_reading(c : Call):
    """
    Creates a Data vertex, and links it to the Handle and Call vertices.
    """
    from ...logger import logger
    from .data import Data, register_read_operation
    if c.status == 1 and c.arguments.file_handle in _inverted_handles:
        logger.debug("Reading from file.")
        h = _inverted_handles[c.arguments.file_handle]
        d = Data()
        d.time = c.time
        d.set_data(c.arguments.buffer)
        f = h.file
        Reads(d, c)
        Conveys(h, d)
        if isinstance(f, File):
            register_read_operation(f, h, d.data, c.arguments.offset)

@chrono
def integrate_file_writing(c : Call):
    """
    Creates a Data vertex, and links it to the Handle and Call vertices.
    """
    from ...logger import logger
    from .data import Data, register_write_operation
    if c.status == 1 and c.arguments.file_handle in _inverted_handles:
        logger.debug("Writing to file.")
        h = _inverted_handles[c.arguments.file_handle]
        d = Data()
        d.time = c.time
        d.set_data(c.arguments.buffer)
        f = h.file
        Writes(c, d)
        Conveys(d, h)
        if isinstance(f, File):
            register_write_operation(f, h, d.data, c.arguments.offset)
    
@chrono
def integrate_file_copying(c : Call):
    """
    Creates a File vertex, and links it to the original file it was copied from.
    """
    from ...logger import logger
    if c.status == 1:
        logger.debug("Copying file.")
        if c.arguments.oldfilepath in _existing_files:
            fs = _existing_files[c.arguments.oldfilepath]
        else:
            fs = File()
            fs.path = c.arguments.oldfilepath
            _existing_files[c.arguments.oldfilepath] = fs
            if c.arguments.oldfilepath not in _all_files:
                _all_files[c.arguments.oldfilepath] = []
            _all_files[c.arguments.oldfilepath].append(fs)
            NewFile(fs).throw()
        if c.arguments.newfilepath in _existing_files:
            fd = _existing_files[c.arguments.newfilepath]
        else:
            fd = File()
            fd.path = c.arguments.newfilepath
            _existing_files[c.arguments.newfilepath] = fd
            if c.arguments.newfilepath not in _all_files:
                _all_files[c.arguments.newfilepath] = []
            _all_files[c.arguments.newfilepath].append(fd)
            NewFile(fd).throw()
        IsCopyiedInto(fs, fd)

@chrono
def integrate_file_deleting(c : Call):
    # You need to make it possible to have multiple times the same file name...
    from ...logger import logger
    from ..config import ColorMap
    if c.status == 1:
        logger.debug("Deleting file.")
        if c.arguments.filepath not in _existing_files:
            f = File()
            f.path = c.arguments.filepath
            _existing_files[c.arguments.filepath] = f
            if c.arguments.filepath not in _all_files:
                _all_files[c.arguments.filepath] = []
            _all_files[c.arguments.filepath].append(f)
            NewFile(f).throw()
        else:
            f = _existing_files[c.arguments.filepath]
        f.color = ColorMap.get_config().Deleted
        _existing_files.pop(c.arguments.filepath)





# File creation
CallHandler(lambda c : c.name == "NtCreateFile", integrate_file_creation)

# File closing
CallHandler(lambda c : c.name == "NtOpenFile", integrate_file_opening)

# File closing
CallHandler(lambda c : c.name == "NtClose", integrate_file_closing)

# File reading
CallHandler(lambda c : c.name == "NtReadFile", integrate_file_reading)

# File writing
CallHandler(lambda c : c.name == "NtWriteFile", integrate_file_writing)

# File copying
CallHandler(lambda c : c.name in {"CopyFileA", "CopyFileExW", "CopyFileW"}, integrate_file_copying)

# File deleting
CallHandler(lambda c : c.name in {"DeleteFileW", "NtDeleteFile"}, integrate_file_deleting)





_N_phase = BuildingPhase.request_finalizing_phase()

def build_fs_tree(e : BuildingPhase):
    """
    When all File nodes have been created, this will replace those which are actually directories into Directory nodes and build the file system tree.
    """
    # TODO : You will need to make a version of this which keeps track of when files/directories exists to make sure your file system tree is indeed a tree !!!
    from ...logger import logger
    from ..graph import find_or_create
    from .network import Host
    from .execution import Runs
    from ..graph import Graph, Edge
    from pathlib import PurePath
    from typing import Type

    disappearing : set[Type[Edge]] = {Runs}


    def mutate(f : File):
        d = Directory()
        d.path = str(f.path)
        d.size = f.size
        graphs = Graph.active_graphs()
        for e in f.edges.copy():
            src, dst = e.source if e.source is not f else d, e.destination if e.destination is not f else d
            e.delete()
            for g in graphs:
                g.remove(e)
            if type(e) in _file_to_dir_cls_table:
                T = _file_to_dir_cls_table[type(e)]
            else:
                T = type(e)
            if T not in disappearing:
                T(src, dst)
        for g in graphs:
            g.remove(f)
            g.append(d)
        return d
    

    if e.major == "Finalizer" and e.minor == _N_phase:
        logger.debug("Mutating up to {} File nodes.".format(len(File)))

        directories : dict[PurePath, list[Directory]] = {}
        files : dict[PurePath, list[File]] = {}
        for filelist in _all_files.values():
            path = filelist[0].path
            for file in filelist:
                paths = [path] + list(path.parents)
                p = paths.pop(0)
                if p not in files:
                    files[p] = []
                files[p].append(file)
                for dpath in paths:
                    if dpath not in directories:
                        directories[dpath] = []
                
        n = 0
        for path in files.copy():
            if path in directories:
                l = list(mutate(file) for file in files.pop(path))
                directories[path].extend(l)
                n += len(l)
        logger.debug("Actually mutated {} File nodes into Directory Nodes.".format(n))
    
        logger.debug("Creating missing directory nodes.")
        for path, filelist in directories.items():
            if not filelist:
                d = Directory()
                d.path = str(path)
                filelist.append(d)
            
        logger.debug("Building file system graph.")
        work : set[File | Directory] = {f for filelist in files.values() for f in filelist}

        while work:
            node = work.pop()
            path = node.path
            parent = path.parent
            if parent != path:
                for parent_node in directories[parent]:
                    Contains(parent_node, node)
                    work.add(parent_node)
            else:
                HasDrive(find_or_create(Host, domain = "host")[0], node)
        
        # Deleting references to irrelevent (mutated) File nodes
        _existing_files.clear()
        _all_files.clear()
                




# TODO : Add a backtracking phase for actual successive copies of files (those without modifications in between)!!!


BuildingPhase.add_callback(build_fs_tree)


del logger, Optional, Color, Arrow, UniqueVertex, Vertex, Call, CallHandler, BuildingPhase, build_fs_tree, chrono, Event, pathlib, Graph, integrate_file_creation, integrate_file_closing, integrate_file_opening, integrate_file_reading, integrate_file_writing, integrate_file_copying, integrate_file_deleting


"""
This module contains the compile function which is used to bake baguettes.
"""

from ..rack import BaguetteRack

__all__ = ["compile"]





def compile(rack : BaguetteRack) -> BaguetteRack:

    """
    Compiles a baguette using the data written in the given BaguetteRack object.
    Returns a BaguetteRack (the same) containing the results (and exceptions on failure).
    """

    from ..rack import BaguetteRack, TimeoutExit

    try:

        if rack.baked:
            return rack
        
        import logging
        levels = {
            0 : logging.ERROR,
            1 : logging.WARNING,
            2 : logging.INFO,
            3 : logging.DEBUG
        }
        
        from .logger import set_level, logger
        set_level(levels[rack.verbosity])

        logger.debug("Just change worker's verbosity level.")

        import os
        from pickle import dump
        from .source.build import Builder
        from .source.config import CompilationParameters
        from .source import types
        from Viper.better_threading import Future, DaemonThread

        if rack.perf:
            from .source.utils import chrono
            chrono.enabled = True
            chrono.auto_report = True

        if not os.path.exists(rack.report):
            raise FileNotFoundError("Could not find report file: {}".format(rack.report))
        if not os.path.isfile(rack.report):
            raise FileExistsError("Given path to report file is not a file.")
        if os.path.exists(rack.baguette) and not os.path.isfile(rack.baguette):
            raise FileExistsError("Given baguette output file exists and is not a file.")
        if os.path.exists(rack.visual) and not os.path.isfile(rack.visual):
            raise FileExistsError("Given visual output file exists and is not a file.")

        if rack.skip_data_comparison:
            CompilationParameters.get_config().SkipLevenshteinForDataNodes = True
        if rack.skip_diff_comparison:
            CompilationParameters.get_config().SkipLevenshteinForDiffNodes = True

        result : Future[bool] = Future()

        def compile_main():
            
            try:
                logger.info("Loading file...")
                with rack.report.open("r") as file:
                    b = Builder(file)
                if not b.data:
                    logger.error("Could not load file properly...")
                    raise RuntimeError("Could not load file properly")
                logger.info("Building graph...")
                b.build()
                G = b.graph
                logger.info("Analyzing graph...")
                logger.info("Got {} vertices and {} edges.".format(G.n, G.m))
                logger.info("Saving with pickle")
                rack.baguette.parent.mkdir(parents = True, exist_ok = True)
                with rack.baguette.open("wb") as file:
                    dump(G, file)
                filters = rack.filters
                if filters:
                    for i, (f, name) in enumerate(zip(filters, rack.filter_names)):
                        logger.info("Applying filters {}/{} : {}".format(i + 1, len(filters), name))
                        f.apply(G)
                logger.info("Exporting to .gexf")
                rack.visual.parent.mkdir(parents = True, exist_ok = True)
                G.export(str(rack.visual))
                result.set(True)
                logger.info("Done !")
            except BaseException as e:
                result.set_exception(e)

        def death_timer():
            from time import sleep
            from Viper.format import duration
            if rack.maxtime < float("inf"):
                logger.info("Death timer thread started. {} remaining.".format(duration(rack.maxtime)))
                sleep(rack.maxtime)
                logger.error("Death timer reached, about to exit.")
                result.set_exception(TimeoutExit("Baking maxtime reached"))
            else:
                while True:
                    sleep(600)
        
        t1 = DaemonThread(target = compile_main)
        t2 = DaemonThread(target = death_timer)
        t1.start()
        t2.start()
        
        while not result.wait(0.1):
            pass
        result.result()
    
    except BaseException as e:
        from traceback import print_exc, TracebackException
        rack.exception = TracebackException.from_exception(e)
        if not isinstance(e, (KeyboardInterrupt, TimeoutExit)):
            print_exc()
    finally:
        if rack.exception is None or not issubclass(rack.exception.exc_type, (KeyboardInterrupt, TimeoutExit)):
            rack.baked = True
        return rack





del BaguetteRack


from io import IOBase
from ..logger import logger
from typing import Dict, Iterator, List, Type, TypeVar
from ..source.utils import *
from ..source.graph import Graph
from ..source.event import Event
import os
from threading import Lock





class BuildingPhase(Event):
    
    """
    This class of event is made to notify the system that we are entering a new buidling phase.
    The major attribute is the name of the phase.
    The minor attribute indicates the iteration number of this phase.
    """

    __lock = Lock()
    __extra_phases = 0

    __slots__ = {
        "major" : "The name of the phase of which the execution is starting.",
        "minor" : "The iteration number of this phase."
    }

    def __init__(self, major : str, minor : int) -> None:
        self.major = major
        self.minor = minor
    
    @staticmethod
    def request_finalizing_phase() -> int:
        """
        Requests a new finalizing phase to be performed by the Builder.
        Returns the integer assigned to the minor of the scheduled finalizer phase.
        """
        with BuildingPhase.__lock:
            n = BuildingPhase.__extra_phases
            BuildingPhase.__extra_phases += 1
            return n
    
    @staticmethod
    def finalizing_steps() -> int:
        """
        Returns the number of finalizing steps tp be performed.
        """
        with BuildingPhase.__lock:
            return BuildingPhase.__extra_phases





class Builder:

    def __init__(self, data : str | IOBase | dict) -> None:
        if isinstance(data, str):
            try:
                data = open(data, "r")
            except:
                raise 
        if isinstance(data, IOBase):
            try:
                from json import load, JSONDecodeError
                data = load(data)
                logger.info("File loaded and structured.")
            except JSONDecodeError:
                data = {}
                logger.error("JSON file has strutural problems.")
        if not isinstance(data, dict):
            raise TypeError("Expected dict or json file, got " + repr(data.__class__.__name__))
        self.data = data
        self._progress = 0
        self._target = 1
    
    @property
    def progress(self) -> float:
        """
        Returns the progress (between 0.0 and 1.0) of the current task.
        """
        return self._progress / self._target

    def calls(self) -> Iterator[Dict[str, Any]]:
        """
        Yields all the calls' data.
        """
        for process in self.data["behavior"]["processes"]:
            for call in process["calls"]:
                yield call

    def machines(self) -> Iterator[Dict[str, str]]:
        """
        Yields all the machines' data.
        """
        if "domains" not in self.data["network"]:
            return
        for machine in self.data["network"]["domains"]:
            yield machine

    def hostIP(self) -> str:
        """
        Returns the host machine's IP address.
        """
        import re
        expr = re.compile(r"\[cuckoo.core.resultserver\] DEBUG: Now tracking machine (\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}) for task #\d+\n")
        for line in self.data["debug"]["cuckoo"]:
            if expr.search(line):
                match = expr.search(line)
                return match.group(1)
        logger.warning("Could not find the host's IP address.")
        return ""
    
    def sample_file_path(self) -> str:
        """
        Finds the name of the sample file in the report and returns its absolute path in the execution environment.
        """
        from ..source.types.execution import Process
        from ..source.utils import parse_command_line
        fname = self.data["target"]["file"]["name"]
        pid = None
        for p in self.data["behavior"]["processtree"]:      # Look only at the processes that Cuckoo started itself.
            for arg in parse_command_line(p["command_line"]):
                if arg.endswith(fname):
                    pid = p["pid"]
        if pid != None:
            for p in Process:
                if p.PID == pid:
                    return p.executable
        raise RuntimeError("Could not find malware process")
        

    @chrono
    def build(self) -> None:
        BuildingPhase("Initialization", 0).throw()
        from ..source.graph import Graph
        from ..source.types.execution import Thread, Call, FollowedBy, Process, NextSignificantCall
        from ..source.types.network import Host, SpawnedProcess
        from ..source.types.filesystem import File, declare_existing_file
        from ..source.graph import find_or_create
        self.graph = Graph()

        with self.graph:
            self.machine : Host = None

            BuildingPhase("Network Discovery", 0).throw()
            logger.info("Identifying machines.")

            for machine in self.machines():
                h = Host()
                h.address = machine["ip"]
                h.domain = machine["domain"]
            
            logger.info("Identifying host machine.")
            
            self.host = find_or_create(Host, address = self.hostIP())[0]
            Host.current = self.host
            self.host.domain = "host"
            self.host.platform = self.data["info"]["platform"]
            self.graph.data["platform"] = self.data["info"]["platform"]

            BuildingPhase("Input Parsing", 0).throw()
            logger.debug("Discovering work on calls.")

            self._target = 0
            for process in self.data["behavior"]["processes"]:
                self._target += len(process["calls"])
            
            N_calls = self._target
            logger.info("{} Calls to find.".format(self._target))

            # Building the basic execution graph
            BuildingPhase("Graph Building", 0).throw()
            logger.info("Building execution star.")
            for process in self.data["behavior"]["processes"]:
                self.graph.add_vertex(self.process_to_vertex(process), False)

            # Creating a file node for the sample file.
            logger.debug("Creating sample file.")
            try:
                path = self.sample_file_path()
                declare_existing_file(path)
            except RuntimeError:
                logger.error("Could not find the process executing the sample.")

            # Discovering calls
            logger.info("Discovering {} calls.".format(N_calls))
            calls : list[Call] = []
            self._target += len(Call) * 2
            seen = set()
            for t in Thread:
                a = t.first
                while a != None:
                    if a in seen:
                        logger.warning("A call has been seen twice while listing.")
                        break
                    seen.add(a)
                    calls.append(a)
                    _target = None
                    for l in a.edges:
                        if isinstance(l, FollowedBy) and l.destination is not a:
                            _target = l.destination
                            break
                    a = _target
            if len(calls) < N_calls:
                logger.warning("{} calls have been forgotten!".format(N_calls - len(calls)))
            
            # Sorting calls based on time
            logger.info("Sorting calls based on time.")
            calls.sort(key = lambda a : a.time)

            # Interpreting each Call
            BuildingPhase("Call Interpretation", 0).throw()
            logger.info("Integrating {} calls.".format(N_calls))
            for a in calls:
                a.integrate()
                self._progress += 1
            
            # Linking processes to host
            BuildingPhase("Process Attribution", 0).throw()
            logger.info("Linking root processes to host machine.")
            p : Process
            for p in Process:
                if not p.parent_process:
                    SpawnedProcess(self.host, p)
            
            # Making skip links
            BuildingPhase("Call Skip-Linking", 0).throw()
            logger.info("Adding skip links in system call sequences")
            for t in Thread:
                last = t.first
                new = last
                while new != None:
                    if new is not last:
                        for l in new.edges:
                            if not isinstance(l, FollowedBy):
                                NextSignificantCall(last, new)
                                last = new
                                break
                    next_call = None
                    for l in new.edges:
                        if isinstance(l, FollowedBy) and l.destination is not new:
                            next_call = l.destination
                            break
                    new = next_call
            
            # Extra building phases
            for i in range(BuildingPhase.finalizing_steps()):
                logger.info("Running finalization phase #{}.".format(i + 1))
                BuildingPhase("Finalizer", i).throw()

    @chrono
    def flush_graph(self):

        logger.info("Flushing graph.")

        from ..source.graph import Vertex, UniqueVertex, Edge, Arrow
        Self = TypeVar("Self")

        def subclass_iterator(cls : Self) -> List[Type[Self]]:
            for c in cls.__subclasses__():
                yield c
                for ci in subclass_iterator(c):
                    yield ci
        
        for cls in subclass_iterator(Vertex):
            if cls != UniqueVertex:
                logger.debug("Adding {} {} vertices to the graph.".format(len(cls), cls.__name__))
                cls.add_vertices_to_graph(self.graph)
        for cls in subclass_iterator(Edge):
            if cls != Arrow:
                logger.debug("Adding {} {} edges to the graph.".format(len(cls), cls.__name__))
                cls.add_edges_to_graph(self.graph)

    @chrono
    def process_to_vertex(self, process : dict) -> "Process":
        from ..source.graph import find_or_create
        from ..source.types.execution import Process, Thread, Call, HasChildProcess, HasThread, HasFirstCall, FollowedBy
        from ..source.types.imports import Import, HasImport
        p : Process
        t : Thread
        a : Call
        p, _ = find_or_create(Process, PID = process["pid"])
        p.executable = process["process_path"]
        p.command = process["command_line"]
        p.start = process["first_seen"]
        p2, _ = find_or_create(Process, PID = process["ppid"])
        e = HasChildProcess(p2, p)

        for imp in process["modules"]:
            basename : str = imp["basename"]
            length : int = imp["imgsize"]
            path : str = imp["filepath"]
            if basename.lower().endswith(".dll"):
                imp, _ = find_or_create(Import, name = basename.lower()[:-4], path = path, length = length)
                HasImport(p, imp)

        for i, ci in enumerate(process["calls"]):
            t, existed = find_or_create(Thread, TID = ci["tid"])
            if not existed:
                HasThread(p, t)
            a = self.api_to_vertex(ci)
            a.thread = t
            if t.first == None:
                t.first = a
                t.last = a
                t.Ncalls = 1
                HasFirstCall(t, a)
            else:
                FollowedBy(t.last, a)
                t.last = a
                t.Ncalls += 1
            self._progress += 1

        for t in p.threads:
            t.start = t.first.time
            t.stop = t.last.time
        
        p.start = min((t.start for t in p.threads), default=p.start)
        p.stop = max((t.stop for t in p.threads), default=p.start)

        return p

    @chrono
    def api_to_vertex(self, call : dict) -> "Call":
        from ..source.record import record
        from ..source.types.execution import Call
        a = Call()
        a.name = call["api"]
        a.category = call["category"]
        a.stacktrace = tuple(call["stacktrace"])
        a.status = call["status"]
        a.arguments = record(**call["arguments"])
        a.return_value = call["return_value"]
        a.time = call["time"]
        a.flags = record(**call["flags"])
        return a





def build(file : str | IOBase) -> "Graph":
    if isinstance(file, str):
        file = open(file, "rb")
    try:
        b = Builder(file)
    except:
        raise TypeError("Expected file, got " + repr(file.__class__.__name__))
    b.build()
    return b.graph


__all__ = ["Builder", "build", "os", "avg", "dev"]